from pyticketswitch.mixins import JSONMixin


class Offer(JSONMixin, object):
    """Describes a reduction in price.

    Attributes:
        seatprice (float): the price per seat/ticket.
        surcharge (float): the additional charges per seat/ticket.
        original_seatprice (float): the original price per seat/ticket.
        original_surcharge (float): the original additional charges per
            seat/ticket.
        absolute_saving (float): the amount of money saved by this offer.
        percentage_saving (float): the amount of money saved by this offer, as
            a percentage of the original price.

    """
    def __init__(self, seatprice=None, surcharge=None, original_seatprice=None,
                 original_surcharge=None, absolute_saving=None,
                 percentage_saving=None):

        self.seatprice = seatprice
        self.surcharge = surcharge
        self.original_surcharge = original_surcharge
        self.original_seatprice = original_seatprice
        self.absolute_saving = absolute_saving
        self.percentage_saving = percentage_saving

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Offer object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns an offer.

        Returns:
            :class:`Offer <pyticketswitch.offer.Offer>`: a new
            :class:`Offer <pyticketswitch.offer.Offer>` object
            populated with the data from the api.

        """
        kwargs = {
            'seatprice': data.get('offer_seatprice'),
            'surcharge': data.get('offer_surcharge'),
            'original_seatprice': data.get('full_seatprice'),
            'original_surcharge': data.get('full_surcharge'),
            'absolute_saving': data.get('absolute_saving'),
            'percentage_saving': data.get('percentage_saving'),
        }
        return cls(**kwargs)



from pyticketswitch.mixins import JSONMixin


class Media(JSONMixin, object):
    """Describes some event media asset

    Attributes:
        caption (str): caption in plain text describing the asset.
        caption_html (str): caption as html describing the asset.
        name (str): name of the asset.
        url (str): url for the asset.
        secure (bool): indicates if the assert url is secure or not.
        width (int): width of the asset in pixels. Only present on the video
        height (int): height of the asset in pixels. Only present on the video
            asset.

    """

    def __init__(self, caption=None, caption_html=None, name=None, url=None,
                 secure=None, width=0, height=0):

        self.caption = caption
        self.caption_html = caption_html
        self.name = name
        self.url = url
        self.secure = secure
        self.width = width
        self.height = height

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Media object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a media asset.

        Returns:
            :class:`Media <pyticketswitch.media.Media>`: a new
            :class:`Media <pyticketswitch.media.Media>` object
            populated with the data from the api.

        """
        url = data.get('secure_complete_url', None)
        secure = True
        if not url:
            url = data.get('insecure_complete_url', None)
            secure = False

        kwargs = {
            'caption': data.get('caption', None),
            'caption_html': data.get('caption_html', None),
            'name': data.get('name', None),
            'url': url,
            'secure': secure,
            'width': data.get('width'),
            'height': data.get('height'),
        }

        return cls(**kwargs)



from pyticketswitch.trolley import Trolley
from pyticketswitch.user import User
from pyticketswitch.utils import isostr_to_datetime
from pyticketswitch.country import Country
from pyticketswitch.customer import Customer
from pyticketswitch.mixins import JSONMixin
from pyticketswitch.address import Address
from pyticketswitch.card_type import CardType
from pyticketswitch.callout import Callout


class Status(JSONMixin, object):
    """Describes the current state of a transaction

    Attributes:
        status (str): the current status of the transaction.
        reserved_at (datetime.datetime): the date and time when the transaction
            was reserved.
        purchased_at (datetime.datetime): the date and time when the transaction
            was purchased.
        trolley (:class:`Trolley <pyticketswitch.trolley.Trolley>`): the
            contents of the transactions trolley.
        external_sale_page (str): the page that was rendered to the customer
            after the transaction was completed. This is only available if
            it was passed into the API at purchase time.
        languages (list): list of IETF language tags relevant to the
            transaction.
        remote_site (str): the remote site the transaction was reserved and
            purchased under.
        unreserved_orders (list): list of :class:`Orders
            <pyticketswitch.order.Order`>` that failed to reserve.
        prefilled_address (:class:`Address <pyticketswitch.address.Address>`):
            some address information that should be used to prefill any
            customer address fields. This is primarily used on B2B accounts.
        needs_payment_card (bool): When :obj:`True` indicates that this
            reservation will require :class:`CardDetails
            <pyticketswitch.payment_methods.CardDetails>` as the payment method
            at purchase time.
        needs_email_address (bool): When :obj:`True` indicates that the
            customer needs to provide a valid email address at purchase time.
        needs_agent_reference (bool): When :obj:`True` indicates that an agent
            reference should be provided at purchase time.
        can_edit_address (bool): When :obj:`False` indicates that the prefilled
            customer address provided by **prefilled_address** should not be
            edited.
        allowed_countries (list): list of :class:`Countries
            <pyticketswitch.country.Country>` that are acceptable for the
            customers postal address to be from.
        minutes_left (float): the number of minutes left before a reservation
            expires.
        supports_billing_address (bool): this transaction will support supplying
            a secondary billing address that is differnet to the customer
            address.
        accepted_cards (list): acceptable debit/credit card types when API
            takes card details from the customer.
        pending_callout (:class:`Callout <pyticketswitch.callout.Callout>`):
            if the transaction is mid purchase then this will hold information
            about how to reenter the purchase process.
    """
    def __init__(self, status=None, reserved_at=None, trolley=None,
                 purchased_at=None, external_sale_page=None,
                 languages=None, remote_site=None, reserve_user=None,
                 prefilled_address=None, needs_payment_card=False,
                 needs_email_address=False, needs_agent_reference=False,
                 can_edit_address=False, allowed_countries=None,
                 minutes_left=None, supports_billing_address=False,
                 accepted_cards=None, pending_callout=None, customer=None,
                 purchase_result=None):

        self.status = status
        self.reserved_at = reserved_at
        self.purchased_at = purchased_at
        self.trolley = trolley
        self.external_sale_page = external_sale_page
        self.languages = languages
        self.remote_site = remote_site
        self.reserve_user = reserve_user
        self.prefilled_address = prefilled_address
        self.needs_payment_card = needs_payment_card
        self.needs_email_address = needs_email_address
        self.needs_agent_reference = needs_agent_reference
        self.can_edit_address = can_edit_address
        self.allowed_countries = allowed_countries
        self.minutes_left = minutes_left
        self.supports_billing_address = supports_billing_address
        self.accepted_cards = accepted_cards
        self.pending_callout = pending_callout
        self.customer = customer
        self.purchase_result = purchase_result

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Status object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a transactions state.

        Returns:
            :class:`Status <pyticketswitch.status.Status>`: a new
            :class:`Status <pyticketswitch.status.Status>` object
            populated with the data from the api.

        """
        accepted_cards = [
            CardType(code=key, description=value)
            for key, value in data.get('accepted_payment_cards', {}).items()
        ]

        customer = None
        customer_api_data = data.get('customer')
        if customer_api_data:
            customer = Customer.from_api_data(customer_api_data)

        kwargs = {
            'status': data.get('transaction_status'),
            'trolley': Trolley.from_api_data(data),
            'remote_site': data.get('remote_site'),
            'can_edit_address': data.get('can_edit_address'),
            'needs_agent_reference': data.get('needs_agent_reference'),
            'needs_email_address': data.get('needs_email_address'),
            'needs_payment_card': data.get('needs_payment_card'),
            'minutes_left': data.get('minutes_left_on_reserve'),
            'supports_billing_address': data.get('supports_billing_address'),
            'accepted_cards': accepted_cards,
            'customer': customer,
        }

        reserved_raw = data.get('reserve_iso8601_date_and_time')
        if reserved_raw:
            kwargs.update(reserved_at=isostr_to_datetime(reserved_raw))

        purchased_raw = data.get('purchase_iso8601_date_and_time')
        if purchased_raw:
            kwargs.update(purchased_at=isostr_to_datetime(purchased_raw))

        external_sale_page_raw = data.get('external_sale_page_raw')
        if external_sale_page_raw:
            raise NotImplementedError("don't know what this looks like yet")

        reserve_user_data = data.get('reserve_user')
        if reserve_user_data:
            reserve_user = User.from_api_data(reserve_user_data)
            kwargs.update(reserve_user=reserve_user)

        languages_raw = data.get('language_list')
        if languages_raw:
            kwargs.update(languages=languages_raw)

        allowed_countries = data.get('allowed_countries')
        if allowed_countries is not None:
            countries = [
                Country(key, description=description)
                for key, description in allowed_countries.items()
            ]
            countries.sort(key=lambda x: x.code)
            kwargs.update(allowed_countries=countries)

        address = data.get('prefilled_address')
        if address:
            kwargs.update(prefilled_address=Address.from_api_data(address))

        pending_callout = data.get('pending_callout')
        if pending_callout:
            kwargs.update(pending_callout=Callout.from_api_data(pending_callout))

        return cls(**kwargs)



from pyticketswitch.mixins import JSONMixin
from pyticketswitch import utils


class Customer (JSONMixin, object):
    """Describes a cutomer to ticketswitch.

    Attributes:
        first_name (str): customers first name.
        last_name (str): customers last name.
        address_lines (list): list of address lines, at least one address line
            is required.
        country_code (str): ISO 3166-1 country code that the customer is living
            in.
        title (str): customer title.
        initials (str): customer initials.
        suffix (str): customer name suffix.
        email (str): email address for contacting the customer.
        post_code (str): post or ZIP code for the customers address.
        town (str): town for the customers address.
        county (str): the county or region of the customers address.
        country (str): the country as a string for the customers address.
        phone (str): a contact phone number for the user.
        home_phone (str): the customers home phone number.
        work_phone (str): the customers work phone number.
        agent_reference (str): agent reference for the customer.
        supplier_can_use_data (bool): indicates that the supplier of the
            tickets is allowed to access and use this data.
        user_can_use_data (bool): indicates that the user (you) can access and
            use this data.
        world_can_use_data (bool): indicates that pretty much anyone is allowed
            to access and use this data.
        first_name_latin (str): **first name** in latin characters.
        last_name_latin (str): **last name** in latin characters.
        address_lines_latin (list): list of :obj:`str` address lines in latin
            characters.
        title_latin (str): **title** in latin characters
        initials_latin (str): **initials** in latin characters.
        suffix_latin (str): **suffix** in latin characters.
        post_code_latin (str): **post_code** in latin characters.
        town_latin (str): **town** in latin characters.
        county_latin (str): **county** in latin characters.
        country_latin (str): **country** in latin characters.
        kwargs (dict): arbitrary keyword arguments to store along with this
            customer object.

    """

    def __init__(self, first_name, last_name, address_lines, country_code,
                 title=None, initials=None, suffix=None, email=None,
                 post_code=None, town=None, county=None, country=None,
                 phone=None, home_phone=None, work_phone=None,
                 agent_reference=None, supplier_can_use_data=False,
                 user_can_use_data=False, world_can_use_data=False,
                 first_name_latin=None, last_name_latin=None,
                 address_lines_latin=None, title_latin=None,
                 initials_latin=None, suffix_latin=None, post_code_latin=None,
                 town_latin=None, county_latin=None, country_latin=None,
                 **kwargs):

        self.first_name = first_name
        self.first_name_latin = first_name_latin

        self.last_name = last_name
        self.last_name_latin = last_name_latin

        self.address_lines = address_lines
        self.address_lines_latin = address_lines_latin

        self.title = title
        self.title_latin = title_latin
        self.initials = initials
        self.initials_latin = initials_latin
        self.suffix = suffix
        self.suffix_latin = suffix_latin

        self.country_code = country_code
        self.post_code = post_code
        self.post_code_latin = post_code_latin
        self.town = town
        self.town_latin = town_latin
        self.county = county
        self.county_latin = county_latin
        self.country = country
        self.country_latin = country_latin

        self.email = email
        self.phone = phone
        self.home_phone = home_phone
        self.work_phone = work_phone

        self.agent_reference = agent_reference

        self.supplier_can_use_data = supplier_can_use_data
        self.user_can_use_data = user_can_use_data
        self.world_can_use_data = world_can_use_data
        self.kwargs = kwargs

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Customer object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a customer.

        Returns:
            :class:`Customer <pyticketswitch.customer.Customer>`: a new
            :class:`Customer <pyticketswitch.customer.Customer>` object
            populated with the data from the api.

        """
        kwargs = {
            'first_name': data.get('first_name'),
            'first_name_latin': data.get('first_name_latin'),
            'last_name': data.get('last_name'),
            'last_name_latin': data.get('last_name_latin'),
            'title': data.get('title'),
            'title_latin': data.get('title_latin'),
            'initials': data.get('initials'),
            'initials_latin': data.get('initials_latin'),
            'suffix': data.get('suffix'),
            'suffix_latin': data.get('suffix_latin'),
            'country_code': data.get('country_code'),
            'post_code': data.get('postcode'),
            'post_code_latin': data.get('postcode_latin'),
            'town': data.get('town'),
            'town_latin': data.get('town_latin'),
            'county': data.get('county'),
            'county_latin': data.get('county_latin'),
            'country': data.get('country'),
            'country_latin': data.get('country_latin'),
            'email': data.get('email_addr'),
            'phone': data.get('phone'),
            'home_phone': data.get('home_phone'),
            'work_phone': data.get('work_phone'),
            'agent_reference': data.get('agent_ref'),
            'supplier_can_use_data': data.get('dp_supplier'),
            'user_can_use_data': data.get('dp_user'),
            'world_can_use_data': data.get('dp_world'),
        }

        address_line_one = data.get('addr_line_one')
        address_line_two = data.get('addr_line_two')

        address_line_one_latin = data.get('addr_line_one_latin')
        address_line_two_latin = data.get('addr_line_two_latin')

        address_lines = []

        if address_line_one:
            address_lines.append(address_line_one)

        if address_line_two:
            address_lines.append(address_line_two)

        address_lines_latin = []

        if address_line_one_latin:
            address_lines_latin.append(address_line_one_latin)

        if address_line_two_latin:
            address_lines_latin.append(address_line_two_latin)

        kwargs.update(
            address_lines=address_lines,
            address_lines_latin=address_lines_latin
        )

        return cls(**kwargs)

    def as_api_parameters(self):
        """Generates a dictionary of parameters to be passed back to the API.

        Returns:
            dict: a set of parameters describing the customer to the API.

        .. note:: the ``latin_*`` attributes are not passed to the API
                   as they are generated internally. As such there is no need
                   to provided them when creating a
                   :class:`Customer <pyticketswitch.customer.Customer>` for
                   API consumption.

        """

        params = {
            'first_name': self.first_name,
            'last_name': self.last_name,
            'country_code': self.country_code,
            'title': self.title,
            'initials': self.initials,
            'suffix': self.suffix,
            'postcode': self.post_code,
            'town': self.town,
            'county': self.county,
            'email_address': self.email,
            'phone': self.phone,
            'work_phone': self.work_phone,
            'home_phone': self.home_phone,
            'agent_ref': self.agent_reference,
            'supplier_can_use_customer_data': self.supplier_can_use_data,
            'user_can_use_customer_data': self.user_can_use_data,
            'world_can_use_customer_data': self.world_can_use_data,
        }

        if self.address_lines and len(self.address_lines) == 1:
            params.update(address_line_one=self.address_lines[0])

        if self.address_lines and len(self.address_lines) > 1:
            params.update(
                address_line_one=self.address_lines[0],
                address_line_two=self.address_lines[1],
            )

        params = utils.filter_none_from_parameters(params)
        return params



import decimal
import requests
import logging
import six
import pyticketswitch
from pyticketswitch import exceptions, utils
from pyticketswitch.event import Event, EventMeta
from pyticketswitch.performance import Performance, PerformanceMeta
from pyticketswitch.availability import AvailabilityMeta
from pyticketswitch.ticket_type import TicketType
from pyticketswitch.send_method import SendMethod
from pyticketswitch.month import Month
from pyticketswitch.discount import Discount
from pyticketswitch.trolley import Trolley
from pyticketswitch.reservation import Reservation
from pyticketswitch.status import Status
from pyticketswitch.user import User
from pyticketswitch.currency import CurrencyMeta
from pyticketswitch.callout import Callout


logger = logging.getLogger(__name__)

POST = 'post'
GET = 'get'
DEFAULT_ROOT_URL = "https://api.ticketswitch.com"


class Client(object):
    """Client wraps the ticketswitch f13 API.

    Client contains auth details and provides helper methods for calling
    the f13 endpoints.

    Attributes:
        user (str): user id used to connect to the api.
        password (str): Password for the user used to connect to the api.
        url (:obj:`str`, optional): Root url for the API.
            Defaults to https://api.ticketswitch.com.
        sub_user (:obj:`str`, optional): the sub user is used to indicate a
            specific agent of the holder of the main users account requests
            is being made for. For example a travel agent might hold the main
            account but an agent in the london office might specify `london` as
            their sub user, whereas an agent in belfast might specify `belfast`
            or even something more specific. Defaults to None
        language (:obj:`str`, optional): prefered IETF language tag. When
            available this will translate text to the specified language. When
            None language defaults to user's preference. Defaults to None.
        tracking_id (:obj:`str`, optional): a tracking ID to use with requests
        use_decimal (bool): parse JSON numbers as decimal. Default is `False`
            but this use is deprecated and decimals are recommended.
        **kwargs: Additional arbitrary key word arguments to keep with the
            object.

    """

    def __init__(self, user, password, url=DEFAULT_ROOT_URL, sub_user=None,
                 language=None, tracking_id=None, use_decimal=False, **kwargs):
        self.user = user
        self.password = password
        self.url = url
        self.sub_user = sub_user
        self.language = language
        self.tracking_id = tracking_id
        self.use_decimal = use_decimal
        self.kwargs = kwargs

    def get_url(self, end_point):
        """Get the url for a given endpoint

        Args:
            end_point (str): the target api end point, for example events.v1 or
                performances.v1

        Returns:
            str: the full url for the given endpoint, contains the base url and
            endpoint

        """

        url = "{url}/f13/{end_point}/".format(
            url=self.url,
            end_point=end_point,
        )
        return url

    @utils.deprecated
    def get_auth_params(self):
        """Get the authentication parameters for inclusion in requests.

        **This method is deprecated** - please use `get_extra_params` instead.

        This this method is intended to be overwritten if
        additional/alternative auth params need to be provided

        Returns:
            dict: auth params that passed to requests

        """
        return {}

    def get_extra_params(self):
        """Get additional parameters for inclusion in requests.

        This method is intended to be overwritten if additional/alternative
        parameters need to be provided in the query string or POST body.

        Returns:
            dict: parameters that will be included in the request
        """
        extra_params = {}
        # Call the deprecated method if it has been overridden
        if not hasattr(self.get_auth_params, 'is_deprecated'):
            utils.deprecation_warning(
                "Function get_auth_params() is deprecated and should not be used",
                stacklevel=3
            )
            extra_params.update(self.get_auth_params())
        elif self.sub_user:
            extra_params.update(sub_id=self.sub_user)

        return extra_params

    def get_auth(self):
        """Get the authentication parameter for the raw request

        This method is intended to be overwritten if required.

        Returns:
            :class:`requests.auth.AuthBase`: the authentication parameter
                accepted by the `requests` module
        """
        if self.user and self.password:
            return (self.user, self.password)

    def get_user_agent(self):
        """Get the user agent for the raw request

        This method is intended to be overwritten by subclasses that want
        a custom User-Agent header for the requests

        Returns:
            :str: the user agent string
        """
        return "pyticketswitch {}".format(pyticketswitch.__version__)

    def get_headers(self, headers):
        """Generate common headers to send with all requests

        Returns:
            dict: key value map of headers

        """
        if self.language:
            headers.update({'Accept-Language': self.language})

        # Modify user agent to report Pyticketswitch version
        headers.update({
            'User-Agent': self.get_user_agent()
        })

        return headers

    def get_tracking_params(self, custom_tracking_id=None):
        """
        Return current request's session tracking id
        or custom tracking id passed
        """
        tracking_id = custom_tracking_id or self.tracking_id
        if tracking_id:
            return {"tsw_session_track_id": tracking_id}
        return {}

    def get_session(self):
        """Get the requests.Session instance to use to make HTTP requests

        By default this method will create a new session for each request. This
        replicates the default behaviour of the requests library:
        https://github.com/kennethreitz/requests/blob/ead8fba84b12e7496c65272a07de47d553aa0ca0/requests/api.py#L57-L58

        A common modification to this class would be to overload this method to
        provide a single session instance across all calls to take advantage of
        keep-alive.

        .. note:: remember to also overload
                  :meth:`cleanup_session <pyticketswitch.client.Client.cleanup_session>`
                  as well or you connections/session might be unexpectedly killed.
        """
        return requests.Session()

    def cleanup_session(self, session):
        """Cleans up sessions so that we don't leave open sockets.

        If you want to add support for persistent connections, then you should
        noop this method so it does nothing.

        Args:
            session (:class:`requests.Session`): the http session to clean up.
        """
        logger.debug('requests session cleaning up')
        session.close()

    def make_request(self, endpoint, params, method=GET, headers={}, timeout=None):
        """Makes actual requests to the API

        Args:
            endpoint (str): target API endpoint
            params (dict): parameters to provide to requests
            method (str): HTTP method to make the request with
                valid values are ``post`` and ``get``. Defaults to ``get``.
            headers (dict): headers to include with the request
            timeout (int): timeout to include with the request. Defaults to ``None``.

        Returns:
            str: The body of the response after deserialising from JSON

        Raises:
            AuthenticationError: When authentication details provided are
                invalid
            InvalidResponseError: When the status code of the response is not
                200
            APIError: When any other explict errors are returned from the API
        """

        url = self.get_url(endpoint)
        params.update(self.get_extra_params())
        if not params.get('tsw_session_track_id'):
            params.update(self.get_tracking_params())

        logger.debug(u'url: %s; endpoint: %s; params: %s', self.url, endpoint, params)

        raw_headers = self.get_headers(headers)

        auth = self.get_auth()

        session = self.get_session()

        if method == POST:
            response = session.post(url, auth=auth, data=params, headers=raw_headers, timeout=timeout)
        else:
            response = session.get(url, auth=auth, params=params, headers=raw_headers, timeout=timeout)

        logger.debug(six.u(response.content))

        self.cleanup_session(session)

        parse_float = decimal.Decimal if self.use_decimal else float

        try:
            contents = response.json(parse_float=parse_float)
        except ValueError:
            raise exceptions.InvalidResponseError(
                ("Unable to parse json data from {} response with status "
                 "code `{}`").format(
                    endpoint,
                    response.status_code,
                )
            )

        if 'error_code' in contents:

            if contents['error_code'] == 3:
                raise exceptions.AuthenticationError(
                    contents['error_desc'],
                    contents['error_code'],
                    response,
                )

            if response.status_code == 410:
                raise exceptions.CallbackGoneError(
                    contents['error_desc'],
                    contents['error_code'],
                    response,
                )

            raise exceptions.APIError(
                contents['error_desc'],
                contents['error_code'],
                response,
            )

        if response.status_code != 200:
            raise exceptions.InvalidResponseError(
                "got status code `{}` from {}".format(
                    response.status_code,
                    endpoint,
                )
            )

        return contents

    def test(self):
        """Test the connection

        Wraps `/f13/test.v1`_

        will return the running user on sucess and will raise an exception
        if the authentication details are invalid.

        This call is not required, but may be useful for validating auth
        credentials, or checking on the health of the ticketswitch API.

        Returns:
            :obj:`pyticketswitch.user.User`: details about the user calling the API

        .. _`/f13/test.v1`: http://docs.ingresso.co.uk/#test

        """
        response = self.make_request('test.v1', {})

        user = User.from_api_data(response)

        return user

    def add_optional_kwargs(self, params, availability=False,
                            availability_with_performances=False,
                            extra_info=False, reviews=False, media=False,
                            cost_range=False, best_value_offer=False,
                            max_saving_offer=False, min_cost_offer=False,
                            top_price_offer=False, no_singles_data=False,
                            cost_range_details=False, source_info=False,
                            tracking_id=None,
                            **kwargs):
        """Adds additional arguments to the requests.

        All client methods will take several optional arguments that will
        extend the data returned with addtional information about the returned
        objects.

        Generally these are applicable to objects that return events, however
        some arguments will also augment performances, cost ranges, and
        availability details.

        `See the main F13 documentation for more details.
        <http://docs.ingresso.co.uk/#additional-parameters>`_

        Args:
            params (dict): The parameters dictionary into which any additional
                paramters will be added
            availability (bool): Includes general information
                about availability of events and performances.
            availability_with_performances (bool): Includes
                detailed information about avilability of events and includes
                performance information.
            extra_info (bool): Includes additional event
                textual content. Defaults to :obj:`False`.
            reviews (bool): Includes event reviews when
                available. Defaults to :obj:`False`.
            media (bool): Includes any media assets associated
                with an event. Defaults to :obj:`False`.
            cost_range (bool): Include price range estimates.
                Defaults to :obj:`False`.
            best_value_offer (bool): requires **cost_range**.
                Includes the offer with the highest percentage saving in cost
                range information. Defaults to :obj:`False`.
            max_saving_offer (bool): requires **cost_range**.
                Includes the offer with the highest absolute saving in cost
                range information. Defaults to :obj:`False`.
            min_cost_offer (bool): requires **cost_range**.
                Includes the offer with the lowest cost in cost range
                information. Defaults to :obj:`False`.
            top_price_offer (bool): requires **cost_range**.
                Includes the offer with the highest cost in cost range
                information. Defaults to :obj:`False`.
            no_singles_data (bool): requires **cost_range**.
                This returns another cost range object that excludes
                availability with only one consecutive seat available.
                Defaults to :obj:`False`.
            cost_range_details (bool): Cost range information for
                each available price band.  Defaults to :obj:`False`.
            source_info (bool): includes information on the source
                system. defaults to :obj:`false`.
            tracking_id (string): Add custom tracking id to the request
            **kwargs: Additional or override parameters to send with the
                request.
        """

        if extra_info:
            params.update(req_extra_info=True)

        if reviews:
            params.update(req_reviews=True)

        if media:
            params.update({
                'req_media_triplet_one': True,
                'req_media_triplet_two': True,
                'req_media_triplet_three': True,
                'req_media_triplet_four': True,
                'req_media_triplet_five': True,
                'req_media_seating_plan': True,
                'req_media_square': True,
                'req_media_landscape': True,
                'req_media_marquee': True,
                'req_video_iframe': True,
            })

        if cost_range:
            params.update(req_cost_range=True)

        if best_value_offer:
            params.update(req_cost_range_best_value_offer=True,
                          req_cost_range=True)

        if max_saving_offer:
            params.update(req_cost_range_max_saving_offer=True,
                          req_cost_range=True)

        if min_cost_offer:
            params.update(req_cost_range_min_cost_offer=True,
                          req_cost_range=True)

        if top_price_offer:
            params.update(req_cost_range_top_price_offer=True,
                          req_cost_range=True)

        if no_singles_data:
            params.update(req_cost_range_no_singles_data=True,
                          req_cost_range=True)

        if cost_range_details:
            params.update(req_cost_range_details=True)

        if availability:
            params.update(req_avail_details=True)

        if availability_with_performances:
            params.update(req_avail_details=True,
                          req_avail_details_with_perfs=True)

        if tracking_id:
            params.update(self.get_tracking_params(
                custom_tracking_id=tracking_id))

        if source_info:
            params.update(req_src_info=True)
        params.update(kwargs)

    def list_events(self, keywords=None, start_date=None, end_date=None,
                    country_code=None, city_code=None, latitude=None,
                    longitude=None, radius=None, include_dead=False,
                    sort_order=None, page=0, page_length=0, **kwargs):

        """List events with the given parameters

        Wraps `/f13/events.v1`_

        Called with out arguments this method will return all live events
        available to the user.

        Args:
            keywords (list): list of keyword strings. For example:
                ``keywords=['sadlers', 'nutcracker']``. Defaults to
                :obj:`None`.
            start_date (datetime.datetime): only show events that
                have a performance after this date. Defaults to :obj:`None`.
            end_date (datetime.datetime): only show events that have
                a performance before this date. Defaults to :obj:`None`.
            country_code (string): only show events in this country.
                the country is specified by it's ISO 3166-1 country code.
                Defaults to :obj:`None`.
            city_code: (string): specified by the internal city code. As
                a rule of thumb this is the lowercase city name followed by
                a hyphen followed by the ISO 3166-1 country code. For example:
                ``london-uk``, ``new_york-us``, ``berlin-de``.  Defaults to
                :obj:`None`.
            latitude (float): only show events near this latitude.
                valid only in combination with longitude and radius. Defaults
                to :obj:`None`.
            longitude (float): only show events near this longitude.
                valid only in combination with latitude and radius. Defaults
                to :obj:`None`.
            radius (float): only show events inside this radius
                relative to the provided coordinates (latitude & longitude
                above). Units are kilometers. Defaults to :obj:`None`.
            include_dead (bool): when true results will include all
                events even if they are dead (are unavailable to purchase
                from). Defaults to :obj:`False`.
            sort_order (string): order the returned events by the
                specified metric. See
                :ref:`sorting search results <sorting_search_results>`.
                Defaults to :obj:`None`
            page (int): the page of a paginated response.
            page_length (int): how many performances are
                returned per page.
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        Returns:
            list, :class:`EventMeta <pyticketswitch.event.EventMeta>`: a list
            of :class:`Events <pyticketswitch.event.Event>` objects and meta
            information for those events.

        Raises:
            InvalidGeoParameters: when latitude, longitude, or radius is
                specified without the rest of the required geographic
                parameters.
            InvalidResponse: when the response is in an unexpected format

        .. _`/f13/events.v1`: http://docs.ingresso.co.uk/#events-list

        """

        params = {}

        if keywords:
            params.update(keywords=','.join(keywords))

        if start_date or end_date:
            params.update(date_range=utils.date_range_str(start_date, end_date))

        if country_code:
            params.update(country_code=country_code)

        if city_code:
            params.update(city_code=city_code)

        if all([latitude, longitude, radius]):
            params.update(circle='{lat}:{lon}:{rad}'.format(
                lat=latitude,
                lon=longitude,
                rad=radius,
            ))
        elif any([latitude, longitude, radius]):
            raise exceptions.InvalidGeoParameters(
                'Geo data must include latitude, longitude, and radius',
            )

        if include_dead:
            params.update(include_dead=True)

        if sort_order:
            params.update(sort_order=sort_order)

        if page > 0:
            params.update(page_no=page)
        if page_length > 0:
            params.update(page_len=page_length)

        self.add_optional_kwargs(params, **kwargs)

        response = self.make_request('events.v1', params)

        if 'results' not in response:
            raise exceptions.InvalidResponseError(
                "got no results key in json response"
            )

        result = response.get('results', {})
        raw_events = result.get('event', [])
        events = [
            Event.from_api_data(data)
            for data in raw_events
        ]

        meta = EventMeta.from_api_data(response)
        return events, meta

    def get_events(self, event_ids,
                   with_addons=False, with_upsells=False, **kwargs):
        """Get events with the given id's

        Wraps `/f13/events_by_id.v1`_

        Args:
            event_ids (list): list of event IDs
            with_addons (bool): include add-on events
            with_upsells (bool): include upsell events
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        Returns:
            dict, :class:`EventMeta <pyticketswitch.event.EventMeta>`:
            :class:`Events <pyticketswitch.event.Event>` indexed by event
            ID and meta information for those events.

        Raises:
            InvalidResponse: when the response is in an unexpected format

        .. _`/f13/events_by_id.v1`: http://docs.ingresso.co.uk/#events-by-id

        """
        params = {}

        if event_ids:
            params.update(event_id_list=','.join(event_ids))

        self.add_optional_kwargs(params, **kwargs)

        if with_addons:
            params.update(add_add_ons=with_addons)

        if with_upsells:
            params.update(add_upsells=with_upsells)

        response = self.make_request('events_by_id.v1', params)

        if 'events_by_id' not in response:
            raise exceptions.InvalidResponseError(
                "got no events_by_id key in json response"
            )

        events_by_id = response.get('events_by_id', {})
        events = {
            event_id: Event.from_events_by_id_api_data(raw_event)
            for event_id, raw_event in events_by_id.items()
            if raw_event.get('event')
        }

        meta = EventMeta.from_api_data(response)
        return events, meta

    def get_event(self, event_id, **kwargs):
        """Get a specific event by id

        Helper method to avoid having to do something like this lots of times::

            >>> client.get_events(['6IF'])['6IF']
            <Event 6IF:Matthew Bourne's Nutcracker TEST>

        Instead this interface is much less stuttery::

            >>> client.get_event('6IF')
            <Event 6IF:Matthew Bourne's Nutcracker TEST>

        Args:
            event_id (str): ID of the event to retrieve
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        Returns:
            :class:`Event <pyticketswitch.event.Event>`, :class:`EventMeta <pyticketswitch.event.EventMeta>`:
            the target event and meta information about the event.

            will return :obj:`None` if the event does not exist.

        """
        events, meta = self.get_events([event_id], **kwargs)
        return events.get(event_id), meta

    def get_months(self, event_id, **kwargs):
        """Returns a summary of availability accross months.

        Wraps `/f13/months.v1`_

        Args:
            event_id (str): ID of the event to retrieve
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        Returns:
            list: :class:`Months <pyticketswitch.month.Month>` ordered chronologically

        Raises:
            InvalidResponse: when the response is in an unexpected format

        .. _`/f13/months.v1`: http://docs.ingresso.co.uk/#months

        """
        params = {'event_id': event_id}

        self.add_optional_kwargs(params, **kwargs)

        response = self.make_request('months.v1', params)

        if 'results' not in response:
            raise exceptions.InvalidResponseError(
                "got no results key in json response"
            )

        result = response.get('results', {})
        raw_months = result.get('month', [])

        months = [
            Month.from_api_data(data)
            for data in raw_months
        ]

        return months

    def list_performances(self, event_id, start_date=None, end_date=None,
                          page_length=0, page=0, **kwargs):
        """List performances for a specified event

        Wraps `/f13/performances.v1`_

        Args:
            event_id (str): identifier for the event.
            start_date (datetime.datetime):
                list performances only if they occur after this date.
            end_date (datetime.datetime):
                list performances only if they occur before this date.
            page (int): the page of a paginated response.
            page_length (int): how many performances are
                returned per page.
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        Returns:
            list, :class:`PerformanceMeta <pyticketswitch.performance.PerformanceMeta>`:
            A list of
            :class:`Performances <pyticketswitch.performance.Performance>`
            for the given event and meta information about the performances.

        Raises:
            InvalidResponse: when the response is in an unexpected format

        .. _`/f13/performances.v1`: http://docs.ingresso.co.uk/#performances-list
        """
        params = {'event_id': event_id}

        if page > 0:
            params.update(page_no=page)

        if page_length > 0:
            params.update(page_len=page_length)

        if start_date or end_date:
            params.update(date_range=utils.date_range_str(start_date, end_date))

        self.add_optional_kwargs(params, **kwargs)

        response = self.make_request('performances.v1', params)

        if 'results' not in response:
            raise exceptions.InvalidResponseError(
                "got no results key in json response"
            )

        result = response.get('results', {})

        raw_performances = result.get('performance', [])
        performances = [
            Performance.from_api_data(data)
            for data in raw_performances
        ]

        meta = PerformanceMeta.from_api_data(response)

        return performances, meta

    def get_performances(self, performance_ids, **kwargs):
        """Get performances with the given ID's

        Wraps `/f13/performances_by_id.v1`_

        Args:
            performance_ids (list): list of performance IDs to fetch.
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        Returns:
            dict, :class:`PerformanceMeta <pyticketswitch.performance.PerformanceMeta>`:
            :class:`Performances <pyticketswitch.performance.Performance>`
            indexed by performance ID and meta information about the
            performances.

        Raises:
            InvalidResponse: when the response is in an unexpected format.

        .. _`/f13/performances_by_id.v1`: http://docs.ingresso.co.uk/#performances-by-id

        """

        params = {
            'perf_id_list': ','.join(performance_ids),
        }

        self.add_optional_kwargs(params, **kwargs)

        response = self.make_request('performances_by_id.v1', params)

        if 'performances_by_id' not in response:
            raise exceptions.InvalidResponseError(
                "got no performances_by_id key in json response"
            )

        raw_performances = response.get('performances_by_id', {})
        performances = {
            performance_id: Performance.from_api_data(data)
            for performance_id, data in raw_performances.items()
        }

        meta = PerformanceMeta.from_api_data(response)

        return performances, meta

    def get_performance(self, performance_id, **kwargs):
        """Get a specific performance by id

        Helper method to avoid having to do something like this lots of times::

            >>> client.get_performances(['6IF-B1H'])['6IF-B1H']
            <Performance 6IF-B1H: 2017-06-02T19:30:00+01:00>

        Instead this interface is much less stuttery::

            >>> client.get_performance('6IF-B1H')
            <Performance 6IF-B1H: 2017-06-02T19:30:00+01:00>

        Args:
            performance_id (str): ID of the performance to retrieve
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        Returns:
            :class:`Performance <pyticketswitch.event.Event>`, :class:`PerformanceMeta <pyticketswitch.performance.PerformanceMeta>`:
            the target Performance and meta information about the performance.

            will return :obj:`None` if the performance does not exist.

        """
        performances, meta = self.get_performances([performance_id], **kwargs)
        return performances.get(performance_id), meta

    def get_availability(self, performance_id, number_of_seats=None,
                         discounts=False, example_seats=False,
                         seat_blocks=False, user_commission=False, **kwargs):
        """Fetch available tickets and prices for a given performance

        Wraps `/f13/availability.v1`_

        Args:
            performance_id (string): identifier of the target performance.
            number_of_seats (int): number of seats that we after,
                Defaults to :obj:`None`.
            discounts (bool): request all discount options for all
                price bands in response. Defaults to :obj:`False`.

            example_seats (bool): request example seats for seated
                events. Defaults to :obj:`False`.
            seat_blocks (bool): request seating information if
                available. Defaults to :obj:`False`.
            user_commission (bool): request user commission for each
                price band/discount. Defaults to :obj:`False`
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        .. note:: setting **discounts** or the **user_commission** arguments to
                  :obj:`True` will increase the response time of your request.

        Returns:
           list, :class:`AvailabilityMeta <pyticketswitch.availability.AvailabilityMeta>`:
           a list of :class:`TicketTypes <pyticketswitch.ticket_type.TicketType>`
           and and meta data about the response.

        Raises:
            BackendBrokenError: when the backend system is responding but
                may be under heavy load
            BackendDownError: when the backend system is not responding
            BackendThrottleError: when your call got throttled and timed out
                while waiting for a response.
            InvalidResponse: when the response is in an unexpected format.

        .. _`/f13/availability.v1`: http://docs.ingresso.co.uk/#availability
        """
        params = {'perf_id': performance_id}

        if number_of_seats:
            params.update(no_of_seats=number_of_seats)

        if discounts:
            params.update(add_discounts=True)

        if example_seats:
            params.update(add_example_seats=True)

        if seat_blocks:
            params.update(add_seat_blocks=True)

        if user_commission:
            params.update(req_predicted_commission=True)

        self.add_optional_kwargs(params, **kwargs)

        response = self.make_request('availability.v1', params)

        if 'availability' not in response:
            raise exceptions.InvalidResponseError(
                "got no availability key in json response"
            )

        meta = AvailabilityMeta.from_api_data(response)

        raw_availability = response.get('availability', {})

        availability = [
            TicketType.from_api_data(data)
            for data in raw_availability.get('ticket_type', [])
        ]

        return availability, meta

    def get_send_methods(self, performance_id, **kwargs):
        """Fetch available delivery methods for a given performance

        Wraps `/f13/send_methods.v1`_

        Args:
            performance_id (string): identifier of the target performance.
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        Returns:
           list, :class:`CurrencyMeta <pyticketswitch.currency.CurrencyMeta>`:
           a list of :class:`SendMethods <pyticketswitch.send_method.SendMethod>`
           and information about the currencies of the prices in the response

        Raises:
            InvalidResponse: when the response is in an unexpected format.

        .. _`/f13/send_methods.v1`: http://docs.ingresso.co.uk/#send-methods
        """
        params = {'perf_id': performance_id}
        self.add_optional_kwargs(params, **kwargs)

        response = self.make_request('send_methods.v1', params)

        if 'send_methods' not in response:
            raise exceptions.InvalidResponseError(
                "got no send_methods key in json response"
            )

        raw_send_methods = response.get('send_methods', {})

        send_methods = [
            SendMethod.from_api_data(data)
            for data in raw_send_methods.get('send_method', [])
        ]

        meta = CurrencyMeta.from_api_data(response)

        return send_methods, meta

    def get_discounts(self, performance_id, ticket_type_code, price_band_code,
                      user_commission=False, **kwargs):
        """Fetch available discounts for a ticket_type/price band combination

        Wraps `/f13/discounts.v1`_

        Args:
            performance_id (string): identifier of the target performance.
            ticket_type_code (string): code for the target ticket type.
            price_band_code (string): code for the target price band.
            user_commission (bool): if True then return the user_commission,
                otherwise do not return the user_commission. Defaults to False.
            **kwargs: see :meth:`add_optional_kwargs <pyticketswitch.client.Client.add_optional_kwargs>`
                for more info.

        Returns:
           list, :class:`CurrencyMeta <pyticketswitch.currency.CurrencyMeta>`:
           a list of :class:`Discounts <pyticketswitch.discount.Discount>`
           and information about the currencies of the prices in the response.

        Raises:
            InvalidResponse: when the response is in an unexpected format.

        .. _`/f13/discounts.v1`: http://docs.ingresso.co.uk/#discounts
        """
        params = {
            'perf_id': performance_id,
            'ticket_type_code': ticket_type_code,
            'price_band_code': price_band_code,
            'req_predicted_commission': user_commission,
        }
        self.add_optional_kwargs(params, **kwargs)

        response = self.make_request('discounts.v1', params)

        if 'discounts' not in response:
            raise exceptions.InvalidResponseError(
                "got no discounts key in json response"
            )

        raw_discounts = response.get('discounts', {})

        discounts = [
            Discount.from_api_data(data)
            for data in raw_discounts.get('discount', [])
        ]

        meta = CurrencyMeta.from_api_data(response)

        return discounts, meta

    def _trolley_params(self, token=None, number_of_seats=None, discounts=None,
                        seats=None, send_codes=None, ticket_type_code=None,
                        performance_id=None, price_band_code=None,
                        item_numbers_to_remove=None, **kwargs):
        """Handle arguments common to the
        :meth:`Client.get_trolley <pyticketswitch.client.Client.get_trolley>`
        and the
        :meth:`Client.make_reservation <pyticketswitch.client.Client.make_reservation>`
        methods.

        These two methods accept identical arguments and these resolve
        to identical parameters.

        Args:
            token (string): trolley token from a previous trolley
                call.
            number_of_seats (int): number of seats to add to the
                trolley.
            discounts (list): list containing discount codes for each
                requested seat.
            seats (list): list of seat IDs.
            send_codes (dict): send codes indexed on backend source
                code.
            ticket_type_code: (string): code of ticket type to add to
                the trolley.
            performance_id: (string): id of the performance to add to
                the trolley.
            price_band_code: (string): code of price band to add to
                the trolley
            item_numbers_to_remove: (list): list of item numbers to
                remove from trolley.
            **kwargs: arbitary additional raw keyword arguments to add the
                parameters.

        Raises:
            InvalidParametersError: when there is an issue with the provided
                parameters.

        """

        params = {}

        if token:
            params.update(trolley_token=token)

        if performance_id:
            params.update(perf_id=performance_id)

        # TODO: check if 0 is a legit number to be passing in here.
        # for the moment I'm assuming that it isn't
        if number_of_seats:
            params.update(no_of_seats=number_of_seats)

        if ticket_type_code:
            params.update(ticket_type_code=ticket_type_code)

        if price_band_code:
            params.update(price_band_code=price_band_code)

        if item_numbers_to_remove and not token:
            raise exceptions.InvalidParametersError(
                'got item_numbers_to_remove but no token specified'
            )
        if item_numbers_to_remove:
            params.update(
                remove_items_list=','.join(
                    [str(item) for item in item_numbers_to_remove]
                )
            )

        if seats:
            params.update({
                'seat{}'.format(i): seat
                for i, seat in enumerate(seats)
            })

        if discounts:
            params.update({
                'disc{}'.format(i): discount
                for i, discount in enumerate(discounts)
            })

        if send_codes and not isinstance(send_codes, dict):
            raise exceptions.InvalidParametersError(
                'send_codes should be a dictionary in the format of `{source_code: send_code}`'
            )

        if send_codes:
            params.update({
                '{}_send_code'.format(source_code): send_code
                for source_code, send_code in send_codes.items()
            })

        self.add_optional_kwargs(params, **kwargs)

        return params

    def get_trolley(self, token=None, number_of_seats=None, discounts=None,
                    seats=None, send_codes=None, ticket_type_code=None,
                    performance_id=None, price_band_code=None,
                    item_numbers_to_remove=None,
                    raise_on_unavailable_order=False, **kwargs):

        """Retrieve the contents of a trolley from the API.

        Wraps `/f13/trolley.v1`_

        Args:
            token (string): trolley token from a previous trolley
                call.
            number_of_seats (int): number of seats to add to the
                trolley.
            discounts (list): list containing discount codes for each
                requested seat.
            seats (list): list of seat IDs.
            send_codes (dict): send codes indexed on backend source
                code.
            ticket_type_code (string): code of ticket type to add to
                the trolley.
            performance_id (string): id of the performance to add to
                the trolley.
            price_band_code (string): code of price band to add to
                the trolley.
            item_numbers_to_remove (list): list of item numbers to
                remove from trolley.
            raise_on_unavailable_order (bool): When set to ``True`` this method
                will raise an exception when the API was not able to add an
                order to the trolley as it was unavailable.
            **kwargs: arbitary additional raw keyword arguments to add the
                parameters.

        Returns:
            :class:`Trolley <pyticketswitch.trolley.Trolley>`, :class:`CurrencyMeta <pyticketswitch.currency.CurrencyMeta>`:
            the contents of the trolley and meta data associated with the
            trolley.

        Raises:
            InvalidParametersError: when there is an issue with the provided
                parameters.
            OrderUnavailableError: when ``raise_on_unavailable_order`` is set
                to ``True`` and the requested addition to a trolley was
                unavailable.

        .. _`/f13/trolley.v1`: http://docs.ingresso.co.uk/#trolley

        """

        params = self._trolley_params(
            token=token, number_of_seats=number_of_seats, discounts=discounts,
            seats=seats, send_codes=send_codes,
            ticket_type_code=ticket_type_code, performance_id=performance_id,
            price_band_code=price_band_code,
            item_numbers_to_remove=item_numbers_to_remove, **kwargs)

        response = self.make_request('trolley.v1', params)

        trolley = Trolley.from_api_data(response)
        meta = CurrencyMeta.from_api_data(response)

        if raise_on_unavailable_order:
            if trolley and trolley.input_contained_unavailable_order:
                raise exceptions.OrderUnavailableError(
                    "inputs contained unavailable order")

        return trolley, meta

    def get_upsells(self, token=None, number_of_seats=None, discounts=None,
                    seats=None, send_codes=None, ticket_type_code=None,
                    performance_id=None, price_band_code=None,
                    item_numbers_to_remove=None, **kwargs):

        """Retrieve a list of upsell events related to a trolley from the API.

        Upsell events are related to a main event but can be purchased
        separately. This call accepts a trolley token or any combination of
        parameters to create a valid trolley, but does not alter any existing
        trolley (it is idempotent).

        Wraps `/f13/upsells.v1`_

        Args:
            token (string): trolley token from a previous trolley
                call.
            number_of_seats (int): trolley with number of seats added.
            discounts (list): list containing discount codes for each
                requested seat.
            seats (list): trolley with added list of seat IDs.
            send_codes (dict): send codes indexed on backend source
                code.
            ticket_type_code: (string): trolley with tickets of
                ticket type added.
            performance_id: (string): trolley with tickets from the
                specified performance added.
            price_band_code: (string): trolley with tickets from a
                specified price band added.
            item_numbers_to_remove: (list): trolley with a list of
                item numbers removed.
            **kwargs: arbitary additional raw keyword arguments to add the
                parameters.

        Returns:
            list, :class:`EventMeta <pyticketswitch.event.EventMeta>`:
            a list of :class:`Events <pyticketswitch.event.Event>` of related upsell
            events

        Raises:
            InvalidParametersError: when there is an issue with the provided
                parameters.
            InvalidResponse: when the response is in an unexpected format.

        .. _`/f13/upsells.v1`: http://docs.ingresso.co.uk/#related-events
        """

        params = self._trolley_params(
            token=token, number_of_seats=number_of_seats, discounts=discounts,
            seats=seats, send_codes=send_codes,
            ticket_type_code=ticket_type_code, performance_id=performance_id,
            price_band_code=price_band_code,
            item_numbers_to_remove=item_numbers_to_remove, **kwargs)

        response = self.make_request('upsells.v1', params)

        if 'results' not in response:
            raise exceptions.InvalidResponseError(
                "got no results key in JSON response"
            )

        results = response.get('results', {})

        raw_upsell_events = results.get('event', [])
        upsell_events = [
            Event.from_api_data(data)
            for data in raw_upsell_events
        ]

        upsell_meta = EventMeta.from_api_data(response)

        return (upsell_events, upsell_meta)

    def get_addons(self, token=None, number_of_seats=None, discounts=None,
                   seats=None, send_codes=None, ticket_type_code=None,
                   performance_id=None, price_band_code=None,
                   item_numbers_to_remove=None, **kwargs):

        """Retrieve a list of add-on events from the API.

        Wraps `/f13/add_ons.v1`_

        Args:
            token (string): trolley token from a previous trolley
                call.
            number_of_seats (int): trolley with number of seats added.
            discounts (list): list containing discount codes for each
                requested seat.
            seats (list): trolley with added list of seat IDs.
            send_codes (dict): send codes indexed on backend source
                code.
            ticket_type_code: (string): trolley with tickets of
                ticket type added.
            performance_id: (string): trolley with tickets from the
                specified performance added.
            price_band_code: (string): trolley with tickets from a
                specified price band added.
            item_numbers_to_remove: (list): trolley with a list of
                item numbers removed.
            **kwargs: arbitrary additional raw keyword arguments to add to the
                parameters.

        Returns:
            list, :class:`EventMeta <pyticketswitch.event.EventMeta>`:
            a list of :class:`Events <pyticketswitch.event.Event>` of add-on events

        Raises:
            InvalidParametersError: when there is an issue with the provided
                parameters.
            InvalidResponse: when the response is in an unexpected format.

        .. _`/f13/add_ons.v1`: http://docs.ingresso.co.uk/#add-ons
        """

        params = self._trolley_params(
            token=token, number_of_seats=number_of_seats, discounts=discounts,
            seats=seats, send_codes=send_codes,
            ticket_type_code=ticket_type_code, performance_id=performance_id,
            price_band_code=price_band_code,
            item_numbers_to_remove=item_numbers_to_remove, **kwargs)

        response = self.make_request('add_ons.v1', params)

        if 'results' not in response:
            raise exceptions.InvalidResponseError(
                "got no results key in json response"
            )

        add_on_results = response.get('results', {})

        raw_add_on_events = add_on_results.get('event', [])
        add_on_events = [
            Event.from_api_data(data)
            for data in raw_add_on_events
        ]

        add_on_meta = EventMeta.from_api_data(response)

        return (add_on_events, add_on_meta)

    def make_reservation(self, token=None, number_of_seats=None, discounts=None,
                         seats=None, send_codes=None, ticket_type_code=None,
                         performance_id=None, price_band_code=None,
                         item_numbers_to_remove=None,
                         raise_on_unavailable_order=False, **kwargs):

        """Attempt to reserve all the items in the given trolley

        Wraps `/f13/reserve.v1`_

        This method will take all the same arguments as
        :func:`get_trolley <pyticketswitch.client.Client.get_trolley>`. You can
        either build a trolley via the
        :func:`get_trolley <pyticketswitch.client.Client.get_trolley>` method
        and pass the returned trolley token into the reservation call, or
        alternatively you can provide the trolley parameters directly to this
        method.

        .. note:: If you are interested in bundling multiple orders into a
                  single purchase then see the :ref:`Bundling <bundling>`
                  documentation for more details.

        Args:
            token (string): trolley token from a previous trolley
                call.
            number_of_seats (int): number of seats to add to the
                trolley.
            discounts (list): list containing discount codes for each
                requested seat.
            seats (list): list of seat id's.
            send_codes (dict): send codes indexed on backend source
                code.
            ticket_type_code: (string): code of ticket type to add to
                the trolley.
            performance_id: (string): id of the performance to add to
                the trolley.
            price_band_code: (string): code of price band to add to
                the trolley
            item_numbers_to_remove: (list): list of item numbers to
                remove from trolley.
            raise_on_unavailable_order (bool): When set to ``True`` this method
                will raise an exception when the API was not able to add an
                order to the trolley as it was unavailable.
            **kwargs: arbitary additional raw keyword arguments to add the
                parameters.

        Returns:
            :class:`Reservation <pyticketswitch.reservation.Reservation>`, :class:`CurrencyMeta <pyticketswitch.currency.CurrencyMeta>`:
            Information about the reservation and meta data asscociated with
            the reservation.

        Raises:
            InvalidParametersError: when there is an issue with the provided
                parameters.
            OrderUnavailableError: when ``raise_on_unavailable_order`` is set
                to ``True`` and the requested addition to a trolley was
                unavailable.

        .. _`/f13/reserve.v1`: http://docs.ingresso.co.uk/#reserve

        """

        params = self._trolley_params(
            token=token, number_of_seats=number_of_seats, discounts=discounts,
            seats=seats, send_codes=send_codes,
            ticket_type_code=ticket_type_code, performance_id=performance_id,
            price_band_code=price_band_code,
            item_numbers_to_remove=item_numbers_to_remove, **kwargs)

        response = self.make_request('reserve.v1', params, method=POST)

        return self.process_reservation_response(response,
                                                 raise_on_unavailable_order)

    def release_reservation(self, transaction_uuid, **kwargs):
        """Release an existing reservation.

        Wraps `/f13/release.v1`_

        Args:
            transaction_uuid (str): the identifier of the reservaiton.
            **kwargs: arbitary additional raw keyword arguments to add the
                parameters.

        Returns:
            bool: :obj:`True` if the reservation was successfully released
            otherwise :obj:`False`.

        .. _`/f13/release.v1`: http://docs.ingresso.co.uk/#release

        """

        params = {'transaction_uuid': transaction_uuid}
        kwargs.update(params)
        response = self.make_request('release.v1', kwargs, method=POST)

        return response.get('released_ok', False)

    def get_reservation(self, transaction_uuid,
                        raise_on_unavailable_order=False, **kwargs):
        """
        This method retrieves a previously made reservation response, verbatim.
        It is faster than the `get_status` method as it is not getting the
        status of the reservation, it is simply retrieving the previously sent
        reservation response, verbatim. Reservation responses are held for 24
        hours, this exceeds the length of time that reservations are valid for.
        Do not interpret this response as meaning that a reservation is still
        being held as it is simply a copy of the original reservation response.

        Wraps `/f13/reserve_page_archive.v1`_

        Args:
            transaction_uuid (string): transaction UUID from a previous
                reservation call.
            raise_on_unavailable_order (bool): When set to ``True`` this method
                will raise an exception when the API was not able to add an
                order to the trolley as it was unavailable.
            **kwargs: arbitary additional raw keyword arguments to add the
                parameters. These will be ignored.

        Returns:
            :class:`Reservation <pyticketswitch.reservation.Reservation>`, :class:`CurrencyMeta <pyticketswitch.currency.CurrencyMeta>`:
            Information about the reservation and meta data asscociated with
            the reservation.

        Raises:
            InvalidParametersError: when there is an issue with the provided
                parameters.
            OrderUnavailableError: when ``raise_on_unavailable_order`` is set
                to ``True`` and the requested addition to a trolley was
                unavailable.

        .. _`/f13/reserve_page_archive.v1`: http://docs.ingresso.co.uk/#reserve_page_archive

        """
        params = {"transaction_uuid": transaction_uuid}
        response = self.make_request('reserve_page_archive.v1', params,
                                     method=GET)

        return self.process_reservation_response(response,
                                                 raise_on_unavailable_order)

    def get_status(self, transaction_uuid=None, transaction_id=None,
                   customer=False, external_sale_page=False, **kwargs):
        """Get the status of reservation, purchase or transaction.

        .. note:: The **transaction_id** field is for backwards compatability
            with the old XML API. You can use either **transaction_id** or
            **transaction_uuid**, however please use the **transaction_uuid** as
            **transaction_id** will be deprecated shortly.

        Wraps `/f13/status.v1`_

        Args:
            transaction_uuid (str): identifier for the transaction.
            transaction_id (str): identifier for the old transaction ids.
                Note: one of these two must be set.
            customer (bool): include customer information if
                available. Defaults to :obj:`False`.
            external_sale_page (bool): include the saved html of the
                sale/confirmation page if you asked us to save it for you.
                Defaults to :obj:`None`
            **kwargs: arbitary keyword parameters to pass directly to the API.

        Returns:
            :class:`Status <pyticketswitch.status.Status>`, :class:`CurrencyMeta <pyticketswitch.currency.CurrencyMeta>`:
            the current status of the transaction and acompanying meta
            information.

        .. _`/f13/status.v1`: http://docs.ingresso.co.uk/#status

        """
        params = {}

        if customer:
            params.update(add_customer=True)

        if external_sale_page:
            params.update(add_external_sale_page=True)

        self.add_optional_kwargs(params, **kwargs)

        if transaction_id:
            params.update(transaction_id=transaction_id)
            response = self.make_request('trans_id_status.v1', params)
        else:
            params.update(transaction_uuid=transaction_uuid)
            response = self.make_request('status.v1', params)

        status = Status.from_api_data(response)
        meta = CurrencyMeta.from_api_data(response)

        return status, meta

    def make_purchase(self, transaction_uuid, customer, payment_method=None,
                      send_confirmation_email=True, agent_reference=None, **kwargs):
        """Purchase tickets for an existing reservation.

        Wraps `/f13/purchase.v1`_

        Args:
            transaction_uuid (str): the identifier of the existing
                reservation/transaction
            customer: `Customer <pyticketswitch.customer.Customer>`
                information.
            payment_method: the customer's payment details. Can be
                :class:`CardDetails <pyticketswitch.payment_methods.CardDetails>`
                or
                :class:`RedirectionDetails <pyticketswitch.payment_methods.RedirectionDetails>`.
            send_confirmation_email (bool): on a successful purchase, when this
                parameter is :obj:`True`, then we will send the customer a
                confirmation email. If you would prefer to send your own
                confirmation email then you can set this parameter to
                :obj:`False`.
            agent_reference (str): an optional custom transaction reference to be
                stored in the Ingresso platform
            **kwargs: arbitrary keyword arguments to send with the request.

        Returns:
            :class:`Status <pyticketswitch.status.Status>`,
            :class:`Callout <pyticketswitch.callout.Callout>`: the current
            status of the transaction and/or a potential callout to redirect
            the customer to.

            This method should only ever return either a status or a callout,
            never both.

            If this method generates a
            :class:`Callout <pyticketswitch.callout.Callout>` then the customer
            should be redirected to the specified third party payment provider.

            See :ref:`Handling callouts <handling_callouts>` for more information.

        .. _`/f13/purchase.v1`: http://docs.ingresso.co.uk/#purchase

        """

        params = {'transaction_uuid': transaction_uuid}

        if send_confirmation_email:
            params.update(send_confirmation_email=True)

        customer_params = customer.as_api_parameters()

        if customer_params:
            params.update(customer_params)

        if payment_method:
            params.update(payment_method.as_api_parameters())

        if agent_reference:
            params.update(agent_reference=agent_reference)

        params.update(kwargs)

        response = self.make_request('purchase.v1', params, method=POST)

        return self.process_purchase_response(response)

    def get_purchase(self, transaction_uuid, **kwargs):
        """
        This method retrieves a previously made purchase response, verbatim.
        It is faster than the `get_status` method as it is not getting the
        status of the purchase, it is simply retrieving the previously sent
        purchase response, verbatim. Purchase responses are held for one year.
        Do not interpret this response as meaning that a purchase has not been
        cancelled as it is simply a copy of the original purchase response.

        Wraps `/f13/purchase_page_archive.v1`_

        Args:
            transaction_uuid (string): transaction UUID from a previous
                purchase call.
            **kwargs: arbitary additional raw keyword arguments to add the
                parameters (these will be ignored).

        Returns:
            :class:`Status <pyticketswitch.status.Status>`,
            :class:`Callout <pyticketswitch.callout.Callout>`: the current
            status of the transaction and/or a potential callout to redirect
            the customer to.

            This method should only ever return either a status or a callout,
            never both.

            If this method generates a
            :class:`Callout <pyticketswitch.callout.Callout>` then the customer
            should be redirected to the specified third party payment provider.

            See :ref:`Handling callouts <handling_callouts>` for more information.

        .. _`/f13/purchase_page_archive.v1`: http://docs.ingresso.co.uk/#purchase_page_archive

        """
        params = {"transaction_uuid": transaction_uuid}
        response = self.make_request('purchase_page_archive.v1', params,
                                     method=GET)

        return self.process_purchase_response(response)

    def next_callout(self, this_token, next_token, returned_data, **kwargs):
        """Gets the next callout in a callout chain.

        Wraps `/f13/callback.v1`_

        At the end of the callout chain the call will return the status of
        the transaction.

        See :ref:`Handling callouts <handling_callouts>` for more information.

        Args:
            this_token (str): the token for the current redirect.
            next_token (str): the token for the potential next redirect.
            returned_data (dict): dictionary of query string paramters appended
                to your return url.
            **kwargs: arbitary keyword arguments to send with the request.

        Returns:
            :class:`Status <pyticketswitch.status.Status>`,
            :class:`Callout <pyticketswitch.callout.Callout>`: the current
            status of the transaction and/or a potential callout to redirect
            the customer to.

            This method should only ever return either a status or a callout,
            never both.

        .. _`/f13/callback.v1`: http://docs.ingresso.co.uk/#purchasing-with-redirect

        """

        endpoint = "callback.v1/this.{}/next.{}".format(this_token, next_token)

        params = returned_data
        params.update(kwargs)

        response = self.make_request(endpoint, params, method=POST)

        callout_data = response.get('callout')

        if callout_data:
            status = None
            callout = Callout.from_api_data(callout_data)
        else:
            callout = None
            status = Status.from_api_data(response)

        meta = CurrencyMeta.from_api_data(response)

        return status, callout, meta

    def process_reservation_response(self, response,
                                     raise_on_unavailable_order):
        reservation = Reservation.from_api_data(response)
        meta = CurrencyMeta.from_api_data(response)

        if raise_on_unavailable_order:
            if reservation and reservation.input_contained_unavailable_order:
                raise exceptions.OrderUnavailableError(
                    "inputs contained unavailable order",
                    reservation=reservation,
                    meta=meta)

        return reservation, meta

    def process_purchase_response(self, response):
        callout_data = response.get('callout')

        if callout_data:
            status = None
            callout = Callout.from_api_data(callout_data)
        else:
            callout = None
            status = Status.from_api_data(response)

        meta = CurrencyMeta.from_api_data(response)

        return status, callout, meta



import six
from pyticketswitch.mixins import JSONMixin


class Currency(JSONMixin, object):
    """Information about the currency prices are described in.

    This information can be used to create human readable prices.

    Attributes:
        code (str): ISO 4217 currency code.
        places (int): precision of decimal numbers.
        pre_symbol (str): a symbol to place before the digits of a price.
        post_symbol (str): a symbol to place after the digits of a price.
        factor (int): arbitary scale factor, can be used to roughly convert
            from one currency to another.
        number (int): internal identifier.

    """

    def __init__(self, code, factor=None, places=None, number=None,
                 pre_symbol=None, post_symbol=None):

        self.factor = factor
        self.places = places
        self.pre_symbol = pre_symbol
        self.post_symbol = post_symbol
        self.code = code
        self.number = number

    def __eq__(self, other):
        """
        FIXME: do we need this anywhere other than tests?
        """
        if not other:
            return False

        if not self.code == other.code:
            return False

        if not self.number == other.number:
            return False

        if not self.factor == other.factor:
            return False

        if not self.places == other.places:
            return False

        if not self.pre_symbol == other.pre_symbol:
            return False

        if not self.post_symbol == other.post_symbol:
            return False

        return True

    def __ne__(self, other):
        return not self == other

    @classmethod
    def from_api_data(cls, data):
        """Convert data from the API into a native object.

        Args:
            data (dict): API data describing a currency.

        Returns:
            :class:`Currency <pyticketswitch.currency.Currency>`: the currency.

        """

        kwargs = {
            'factor': data.get('currency_factor'),
            'places': data.get('currency_places'),
            'number': data.get('currency_number'),
            'pre_symbol': data.get('currency_pre_symbol'),
            'post_symbol': data.get('currency_post_symbol'),
        }

        return cls(data.get('currency_code'), **kwargs)

    def price_as_string(self, price):
        """Generates a human readble string for a price.

        Args:
            price (float): a price

        Returns:
            str: the price with the correct number of places with pre and post
            symbols attached

        Example:

            >>> from pyticketswitch.currency import Currency
            >>> usd = Currency('usd', pre_symbol='$', places=2)
            >>> usd.price_as_string(5)
            u'$5.00'
            >>> usd.price_as_string(12.34567)
            u'$12.34'

        """
        price = price if price else 0
        format_string = six.text_type('{pre}{price:.' + str(self.places) + 'f}{post}')
        return format_string.format(
            pre=self.pre_symbol or '',
            price=price,
            post=self.post_symbol or '',
        )

    def __repr__(self):
        return u'<Currency {}>'.format(self.code)


class CurrencyMeta(JSONMixin, object):
    """Currency information for another object.

    Attributes:
        currencies (dict) dictionary of
            :class:`Currency <pytickectswitch.currency.Currency>`) objects
            indexed on currency code.
        default_currency_code (str): unless other wise specified all prices in
            the related response will be in this currency.
        desired_currency_code (str):
            the currency that the user account is expecting. Useful for
            conversions.

    """

    def __init__(self, currencies, default_currency_code=None,
                 desired_currency_code=None):
        self.currencies = currencies
        self.default_currency_code = default_currency_code
        self.desired_currency_code = desired_currency_code

    @classmethod
    def from_api_data(cls, data):
        """Convert data from the API into a native object.

        Args:
            data (dict): API data describing the currencies of a related
                object.

        Returns:
            :class:`CurrencyMeta <pyticketswitch.currency.CurrencyMeta>`: the
                currency meta data.

        """
        currency_data = data.get('currency_details')
        if not currency_data:
            return cls(currencies={})

        currencies = {
            code: Currency.from_api_data(value)
            for code, value in currency_data.items()
        }

        default_currency_code = data.get('currency_code')
        desired_currency_code = data.get('desired_currency_code')

        return cls(
            currencies=currencies,
            default_currency_code=default_currency_code,
            desired_currency_code=desired_currency_code,
        )

    def get_currency(self, code=None):
        """Get the :class:`Currency <pyticketswitch.currency.Currency>` object for the given code.

        Args:
            code (str, optional): ISO 4217 currency code.

        Returns:
            :class:`Currency <pyticketswitch.currency.Currency>`: The currency
            for the given code.

            If no code is specified, return the default currency as specified
            by :attr:`CurrencyMeta.default_currency_code <pyticketswitch.currency.CurrencyMeta.default_currency_code>`.

            If no default currency is specified return :obj:`None`.

        """
        if not code:
            code = self.default_currency_code

        if not code:
            return None

        return self.currencies.get(code)

    def get_desired_currency(self):
        """Get the :class:`Currency <pyticketswitch.currency.Currency>` object for the desired currency.

        Returns:
            :class:`Currency <pyticketswitch.currency.Currency>`: The currency
            for the desired currency.

            If no desired currency is specified return :obj:`None`.
        """
        if not self.desired_currency_code:
            return None

        return self.currencies.get(self.desired_currency_code)



from pyticketswitch.mixins import JSONMixin


class Field(JSONMixin, object):
    """Describes a custom field for an event

    Attributes:
        name (str): the name of the field.
        label (str): human readble name of the field.
        data (str): the field data.

    """

    def __init__(self, name, label, data):
        self.name = name
        self.label = label
        self.data = data

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Field object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a field.

        Returns:
            :class:`Field <pyticketswitch.field.Field>`: a new
            :class:`Field <pyticketswitch.field.Field>` object
            populated with the data from the api.

        """

        kwargs = {
            'name': data.get('custom_field_name'),
            'label': data.get('custom_field_label'),
            'data': data.get('custom_field_data'),
        }

        return cls(**kwargs)



import datetime
from pyticketswitch.mixins import JSONMixin
from pyticketswitch.currency import CurrencyMeta
from pyticketswitch.misc import MONTH_NUMBERS


class AvailabilityMeta(CurrencyMeta):
    """Meta data about an availability response

    Attributes:
        contiguous_seat_selection_only (bool): indicates that the backend
            system will only allow seats to be selected that are in a
            contiguous line.
        valid_quantities (list): list of valid number of tickets available for
            selection.
        max_bundle_size (int): maximum bundle size for this event. If it is
            `None` then there is no maximum size.
        currencies (dict): dictionary of
            :class:`Currency <pytickectswitch.currency.Currency>`) objects
            indexed on currency code.
        default_currency_code (str): unless other wise specified all prices in
            the related response will be in this currency.
        desired_currency_code (str):
            the currency that the user account is expecting. Useful for
            conversions.
        backend_is_broken (bool): indicates that the backend system is presently
            marked as broken.
        backend_is_down (bool): indicates that we are not currently able to
            contact the backend system.
        backend_throttle_failed (bool): indicates that your call was throttled
            and failed to get a slot inside a viable timeframe. When true this
            indicates that the backend system is under heavy load.

    """
    def __init__(self, contiguous_seat_selection_only=True, currency=None,
                 valid_quantities=None, max_bundle_size=None, backend_is_down=False,
                 backend_is_broken=False, backend_throttle_failed=False,
                 *args, **kwargs):
        self.contiguous_seat_selection_only = contiguous_seat_selection_only
        self.currency = currency
        self.max_bundle_size = max_bundle_size
        self.valid_quantities = valid_quantities
        self.backend_is_broken = backend_is_broken
        self.backend_is_down = backend_is_down
        self.backend_throttle_failed = backend_throttle_failed
        super(AvailabilityMeta, self).__init__(*args, **kwargs)

    @classmethod
    def from_api_data(cls, data):
        """Creates a new AvailabilityMeta object from API data from ticketswitch.

        Args:
            data (dict): the whole response from the availability call.

        Returns:
            :class:`AvailabilityMeta <pyticketswitch.availability.AvailabilityMeta>`:
            a new
            :class:`AvailabilityMeta <pyticketswitch.availability.AvailabilityMeta>`
            object populated with the data from the api.

        """
        inst = super(AvailabilityMeta, cls).from_api_data(data)

        inst.valid_quantities = data.get('valid_quantities')
        inst.max_bundle_size = data.get('max_bundle_size')
        inst.contiguous_seat_selection_only = data.get('contiguous_seat_selection_only', True)
        inst.backend_is_broken = data.get('backend_is_broken')
        inst.backend_is_down = data.get('backend_is_down')
        inst.backend_throttle_failed = data.get('backend_throttle_failed')

        return inst


class AvailabilityDetails(JSONMixin, object):
    """Details about the available tickets of a performance.

    Attributes:
        ticket_type_code (str): identifier of the ticket type.
        ticket_type_description (str): human readable description of the ticket
            type.
        price_band_code (str): identifier of the price band.
        price_band_description (str): human readable description of the price
            band.
        seatprice (float): price of an individual seat.
        surcharge (float): additional charges per seat.
        full_seatprice (float): the non-offer price of an individual seat.
        full_surcharge (float): the non-offer additional charges per seat.
        currency (str): the currency of the prices.
        first_date (datetime.datetime): the first date and time this
            combination of ticket type and price band is available from.
        last_date (datetime.datetime): the latest date and time this
            combination of ticket type and price band is available from.
        valid_quantities (list): list of valid number of tickets available for
            selection.
        cached_number_available (int): the maximum number of consecutive
            tickets available.

    """

    def __init__(self, ticket_type=None, price_band=None,
                 ticket_type_description=None, price_band_description=None,
                 seatprice=None, surcharge=None, full_seatprice=None,
                 full_surcharge=None, currency=None, first_date=None,
                 percentage_saving=None, absolute_saving=None,
                 last_date=None, calendar_masks=None,
                 weekday_mask=None, valid_quantities=None,
                 cached_number_available=None):

        self.ticket_type = ticket_type
        self.ticket_type_description = ticket_type_description
        self.price_band = price_band
        self.price_band_description = price_band_description
        self.seatprice = seatprice
        self.surcharge = surcharge
        self.full_seatprice = full_seatprice
        self.full_surcharge = full_surcharge
        self.percentage_saving = percentage_saving
        self.absolute_saving = absolute_saving
        self.currency = currency
        self.first_date = first_date
        self.last_date = last_date
        self._calendar_masks = calendar_masks
        self._weekday_mask = weekday_mask
        self.valid_quantities = valid_quantities
        self.cached_number_available = cached_number_available

    @classmethod
    def from_api_data(cls, data):
        """Creates AvailabilityDetails from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns availability details.

        Returns:
            list: a list of
            :class:`AvailabilityDetails <pyticketswitch.availability.AvailabilityDetails>`
            objects populated with the data from the api.

        """
        details = []

        for ticket_type in data.get('ticket_type', []):
            kwargs = {
                'ticket_type': ticket_type.get('ticket_type_code'),
                'ticket_type_description': ticket_type.get('ticket_type_desc'),
            }

            for price_band in ticket_type.get('price_band', []):
                kwargs.update(
                    price_band=price_band.get('price_band_code'),
                    price_band_description=price_band.get('price_band_desc'),
                )

                raw_details_list = price_band.get('avail_detail', [{}])

                for raw_details in raw_details_list:

                    kwargs['seatprice'] = raw_details.get('seatprice', 0.0)
                    kwargs['full_seatprice'] = raw_details.get('full_seatprice', 0.0)
                    kwargs['surcharge'] = raw_details.get('surcharge', 0.0)
                    kwargs['full_surcharge'] = raw_details.get('full_surcharge', 0.0)
                    kwargs['percentage_saving'] = raw_details.get('percentage_saving', 0.0)
                    kwargs['absolute_saving'] = raw_details.get('absolute_saving', 0.0)
                    kwargs['currency'] = raw_details.get('avail_currency_code')

                    available_dates = raw_details.get('available_dates', {})
                    if 'first_yyyymmdd' in available_dates:
                        try:
                            first = datetime.datetime.strptime(
                                available_dates['first_yyyymmdd'],
                                '%Y%m%d'
                            )
                            kwargs['first_date'] = first.date()
                        except ValueError:
                            pass

                    if 'last_yyyymmdd' in available_dates:
                        try:
                            last = datetime.datetime.strptime(
                                available_dates['last_yyyymmdd'],
                                '%Y%m%d'
                            )
                            kwargs['last_date'] = last.date()
                        except ValueError:
                            pass

                    kwargs['calendar_masks'] = {
                        int(year[5:]): {
                            MONTH_NUMBERS[month[:3]]: mask
                            for month, mask in month_masks.items()
                        }
                        for year, month_masks in available_dates.items()
                        if year.startswith('year_')
                    }

                    if 'available_weekdays_bitmask' in raw_details:
                        kwargs['weekday_mask'] = raw_details['available_weekdays_bitmask']

                    if 'valid_quantities' in raw_details:
                        kwargs['valid_quantities'] = raw_details['valid_quantities']

                    kwargs['cached_number_available'] = (
                        raw_details.get('cached_number_available'))

                    avail_details = AvailabilityDetails(**kwargs)

                    if avail_details._weekday_mask:
                        weekday_list = []
                        for day in range(0, 7):
                            weekday_list.append(avail_details.on_weekday(day))
                        avail_details.weekday_list = weekday_list
                    details.append(avail_details)

        return sorted(details, key=lambda x: (x.ticket_type_description, x.combined_price()))

    def is_available(self, year, month=None, day=None):
        """Check availablity on a given year/month/day.

        Args:
            year (int): the year to check. For example, 2005, 2016, 2222, etc
            month (int): (optional) the month to check. 1 == jan, 12 == dec, etc
            day (int): (optional) the day of the month, zero indexed.

        Returns:
            bool: :obj:`True` when availble inside the given parameters, other
            wise :obj:`False`.

        Raises:
            ValueError: when a day is specified but not the month.

        """
        if day and not month:
            raise ValueError('a month must be specified to specify a day')

        if year not in self._calendar_masks:
            return False

        year_data = self._calendar_masks[year]

        if month and month not in year_data:
            return False

        month_mask = year_data.get(month)

        # the month mask is a 32 bit int where the right most bit is the first
        # day of the month. To find if we have availability on a specfic day,
        # we bit shift the mask by the zero indexed day and AND it by 1. If the
        # result is a 1 then this combo of ticket type and price band is
        # available on that day, otherwise it is not.
        if day and (month_mask >> (day - 1) & 1) == 0:
            return False

        return True

    def on_weekday(self, day):
        """
        Check if this combination of ticket_type and price band is available
        on a given day of the week, where 0 is monday and sunday is 6.

        .. note:: the api uses sunday as day 0 where as python uses monday.
                  I've made a concious decision here to use the python numbers
                  to keep it consistant with anything written against this.
                  (also sunday as day 0 is silly)

        Args:
            day: the day of the week zero indexed from monday.

        Returns:
            bool: :obj:`True` when available on the given day of the week,
            otherwise :obj:`False`.

        """

        adjusted_day = day + 1 if day < 6 else 0

        return bool(self._weekday_mask >> adjusted_day & 1)

    def combined_price(self):
        return self.seatprice + self.surcharge

    def combined_full_price(self):
        return self.full_seatprice + self.full_surcharge



from pyticketswitch.exceptions import IntegrityError
from pyticketswitch.cost_range import CostRange
from pyticketswitch.ticket_type import TicketType
from pyticketswitch.content import Content
from pyticketswitch.media import Media
from pyticketswitch.review import Review
from pyticketswitch.availability import AvailabilityDetails
from pyticketswitch.field import Field
from pyticketswitch.mixins import JSONMixin, PaginationMixin
from pyticketswitch.currency import CurrencyMeta


class Event(JSONMixin, object):
    """Describes a product in the ticketswitch system.

    Attributes:
        id (str): the identifier for the event.
        status (str): status of the event.
        description (str): human readable name for the event.
        source (str): the backend system from wich the event originates.
        source_code (str): the internal code for the backend system.
        event_type (str): the type of the event.
        venue (str): a human readable description of the venue.
        classes (dict): a dictionary of class descriptions that the event
            belongs to keyed on class identifier.
        filters (list): a list of filters that the event belongs to.
        postcode (str): venue post code.
        city (str): human readable venue city.
        city_code (str): venue city code
        country (str): human readable country name.
        country_code (str): ISO 3166-1 country code.
        latitude (float): latitude of the venue.
        longitude (float): longitude of the venue.
        max_running_time (int): maximum running time of a performance in
            minutes.
        min_running_time (int): minimum running time of a performance in
            minutes.
        show_performance_time (bool): indicates that the performance time
            for this event is relevant and should be shown.
        has_performances (bool): indicates that the event has performances.
        is_seated (bool): indicates the event is seated.
        needs_departure_date (bool): indicates that ticket purchases for this
            event will require a departure date.
        needs_duration (bool): indicates that ticket purchases for this
            event will require a duration.
        needs_performance (bool): indicates that you should ask your customer
            what date/time they want to attend this event on.
            See :ref:`Events that don't need performance selection
            <events_that_dont_need_performance_selection>` for more
            information.
        upsell_list (list): list of event id's.
        cost_range (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            pricing summary from cached availability. Only present when
            requested.
        no_singles_cost_range (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            pricining summary from cached availability. Only present when
            requested.
        cost_range_details (:class:`CostRangeDetails <pyticketswitch.cost_range.CostRangeDetails>`):
            summary pricing information broken down by availability. This is
            cached data. Only present when requested.
        content (dict): dictionary of
            :class:`Content <pyticketswitch.content.Content>` objects
            indexed on content name. Only present when requested.
        fields (dict): dictionary of
            :class:`Field <pyticketswitch.content.Content>` objects
            indexed on field name. Only present when requested.
        event_info (str): event info in plain text. Only present when
            requested.
        event_info_html (str): event info as HTML. Only present when requested.
        venue_addr (str): venue address in plain text. Only present when
            requested.
        venue_addr_html (str): venue address as HTML. Only present when requested.
        venue_info (str): venue info in plain text. Only present when
            requested.
        venue_info_html (str): venue info as HTML. Only present when requested.
        media (dict): dictionary of :class:`Media <pyticketswitch.media.Media>`
            objects indexed on media name. Only present when requested.
        reviews (list): list of :class:`Reviews <pyticketswitch.review.Review>`
            objects. Only present when requested.
        critic_review_percent (float): summary of critic review star rating.
            rated from 1 (lowest) to 5 (highest).
        availability_details (:class:`AvailabilityDetails <pyticketswitch.availability.AvailabilityDetails>`):
            summary of availability details from cached data. Only present when
            requested.
        component_events (list): list of :class:`Events <pyticketswitch.event.Event>`
            objects that comprise a meta event.
        valid_quantities (list): list of valid quanities available for
            purchase. from cached data, only available when requested by
            **get_events** or **get_event**.
        raw (dict): the raw data used to generate the object.
        is_addon (bool): indicates that the event is an addon.
        is_auto_quantity_add_on (bool): Indicates whether add on quantity will
            be modified based on the number of ticket orders, if true number
            of add ons will be equal to total number of tickets for all
            parent events in the trolley.
        area_code (str): the internal code for the area. This is for internal
            use only.
        venue_code (str): the internal code for the venue. This is for internal
            use only.
        venue_is_enforced (bool): indicates if the venue is enforced. This is
            for internal use only.
        lingo_code (str): a code for the type of event, e.g. theatre or
            attraction. This is for internal use only.
    """

    def __init__(self, id_, status=None, event_type=None, source=None,
                 source_code=None, venue=None, description=None, postcode=None,
                 classes=None, filters=None, upsell_list=None, city=None,
                 city_code=None, country=None, country_code=None,
                 latitude=None, longitude=None, needs_departure_date=False,
                 needs_duration=False, addon_events=None, upsell_events=None,
                 needs_performance=False, has_performances=False,
                 is_seated=False, show_performance_time=False,
                 min_running_time=None, max_running_time=None, cost_range=None,
                 no_singles_cost_range=None, cost_range_details=None,
                 content=None, event_info_html=None, event_info=None,
                 venue_addr_html=None, venue_addr=None, venue_info=None,
                 venue_info_html=None, media=None, reviews=None,
                 critic_review_percent=None, availability_details=None,
                 component_events=None, valid_quantities=None, fields=None,
                 raw=None, is_add_on=False, area_code=None,
                 venue_code=None, venue_is_enforced=None,
                 lingo_code=None, is_auto_quantity_add_on=False):

        self.id = id_
        self.status = status
        self.description = description
        self.source = source
        self.source_code = source_code
        self.event_type = event_type
        self.venue = venue

        self.classes = classes
        self.filters = filters

        self.postcode = postcode
        self.city = city
        self.city_code = city_code
        self.country = country
        self.country_code = country_code
        self.latitude = latitude
        self.longitude = longitude

        self.max_running_time = max_running_time
        self.min_running_time = min_running_time

        self.show_performance_time = show_performance_time
        self.has_performances = has_performances
        self.is_seated = is_seated
        self.needs_departure_date = needs_departure_date
        self.needs_duration = needs_duration
        self.needs_performance = needs_performance

        self.addon_events = addon_events
        self.upsell_events = upsell_events
        # TODO This will be removed from the API soon
        self.upsell_list = upsell_list

        self.cost_range = cost_range
        self.no_singles_cost_range = no_singles_cost_range
        self.cost_range_details = cost_range_details

        self.content = content
        self.fields = fields
        self.event_info = event_info
        self.event_info_html = event_info_html
        self.venue_addr = venue_addr
        self.venue_addr_html = venue_addr_html
        self.venue_info = venue_info
        self.venue_info_html = venue_info_html

        self.media = media
        self.reviews = reviews
        self.critic_review_percent = critic_review_percent

        self.availability_details = availability_details

        self.component_events = component_events

        self.valid_quantities = valid_quantities
        self.raw = raw

        self.is_add_on = is_add_on
        self.is_auto_quantity_add_on = is_auto_quantity_add_on

        self.venue_code = venue_code
        self.area_code = area_code
        self.venue_is_enforced = venue_is_enforced
        self.lingo_code = lingo_code

    @classmethod
    def class_dict_from_api_data(cls, data):
        """Creates a dict of Event data from a raw ticketswitch API call

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a event.

        Returns:
            dict: a new dict populated with the data from the api for creating
            an :class:`Event <pyticketswitch.event.Event>` object

        """

        id_ = data.get('event_id')

        if not id_:
            raise IntegrityError("event_id not found in event data", data=data)

        geo_data = data.get('geo_data', {})

        # the raw field 'has_no_perfs' is a negative flag, so I'm inverting it
        has_performances = not data.get('has_no_perfs', False)

        api_cost_range = data.get('cost_range', {})
        api_no_singles_cost_range = api_cost_range.get('no_singles_cost_range', {})
        cost_range = None
        no_singles_cost_range = None

        if api_cost_range:
            api_cost_range['singles'] = True
            cost_range = CostRange.from_api_data(api_cost_range)

        if api_no_singles_cost_range:
            api_no_singles_cost_range['singles'] = False
            no_singles_cost_range = CostRange.from_api_data(
                api_no_singles_cost_range)

        api_cost_range_details = data.get('cost_range_details', {})
        ticket_type_list = api_cost_range_details.get('ticket_type', [])
        cost_range_details = [
            TicketType.from_api_data(ticket_type)
            for ticket_type in ticket_type_list
        ]

        api_content = data.get('structured_info', {})
        content = {
            key: Content.from_api_data(value)
            for key, value in api_content.items()
        }

        fields = {
            field.get('custom_field_name'): Field.from_api_data(field)
            for field in data.get('custom_fields', {})
        }

        media = {}
        api_media = data.get('media', {})
        for asset in api_media.get('media_asset', []):
            new_media = Media.from_api_data(asset)
            media[new_media.name] = new_media

        api_video = data.get('video_iframe')
        if api_video:
            kwargs = {
                'secure_complete_url': api_video.get('video_iframe_url_when_secure'),
                'insecure_complete_url': api_video.get('video_iframe_url_when_insecure'),
                'caption': api_video.get('video_iframe_caption'),
                'caption_html': api_video.get('video_iframe_caption_html'),
                'width': api_video.get('video_iframe_width'),
                'height': api_video.get('video_iframe_height'),
                'name': 'video',
            }
            new_video = Media.from_api_data(kwargs)
            media['video'] = new_video

        api_reviews = data.get('reviews', {})
        reviews = [
            Review.from_api_data(api_review)
            for api_review in api_reviews.get('review', [])
        ]

        availability_details = AvailabilityDetails.from_api_data(
            data.get('avail_details', {}))
        api_component_events = data.get('meta_event_component_events', {})
        component_events = [
            Event.from_api_data(meta_event)
            for meta_event in api_component_events.get('event', [])
        ]

        lingo_code = None
        raw_lingo_data = data.get('lingo_data')
        if raw_lingo_data:
            lingo_code = raw_lingo_data.get('lingo_code')

        kwargs = {
            'id_': id_,
            'description': data.get('event_desc', None),
            'status': data.get('event_status'),
            'event_type': data.get('event_type'),
            'source_code': data.get('source_code'),
            'source': data.get('source_desc'),
            'venue': data.get('venue_desc'),

            'classes': data.get('classes'),
            #TODO: don't actually know what filters look like yet...
            'filters': data.get('custom_filter', []),
            'fields': fields,

            'postcode': data.get('postcode'),
            'city': data.get('city_desc'),
            'city_code': data.get('city_code'),
            'country': data.get('country_desc'),
            'country_code': data.get('country_code'),

            'latitude': geo_data.get('latitude'),
            'longitude': geo_data.get('longitude'),

            'max_running_time': data.get('max_running_time'),
            'min_running_time': data.get('min_running_time'),

            'has_performances': has_performances,
            'show_performance_time': data.get('show_perf_time', False),
            'is_seated': data.get('is_seated', False),
            'needs_departure_date': data.get('need_departure_date', False),
            'needs_duration': data.get('need_duration', False),
            'needs_performance': data.get('need_performance', False),

            'upsell_list': data.get('event_upsell_list', {}).get('event_id', []),

            'cost_range': cost_range,
            'no_singles_cost_range': no_singles_cost_range,
            'cost_range_details': cost_range_details,

            # extra info
            'event_info_html': data.get('event_info_html'),
            'event_info': data.get('event_info'),
            'venue_addr_html': data.get('venue_addr_html'),
            'venue_addr': data.get('venue_addr'),
            'venue_info': data.get('venue_info'),
            'venue_info_html': data.get('venue_info_html'),
            'content': content,

            'media': media,
            'reviews': reviews,
            'critic_review_percent': data.get('critic_review_percent'),

            'availability_details': availability_details,
            'component_events': component_events,
            'valid_quantities': data.get('valid_quantities'),
            'raw': data,

            'is_add_on': data.get('is_add_on', False),
            'is_auto_quantity_add_on': data.get('is_auto_quantity_add_on', False),

            'venue_code': data.get('venue_code'),
            'area_code': data.get('area_code'),
            'lingo_code': lingo_code,
        }

        return kwargs

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Event object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a event.

        Returns:
            :class:`Event <pyticketswitch.event.Event>`: a new
            :class:`Event <pyticketswitch.event.Event>` object
            populated with the data from the api.

        """

        kwargs = cls.class_dict_from_api_data(data)
        return cls(**kwargs)

    @classmethod
    def from_events_by_id_api_data(cls, data):
        """Creates a new Event object from API data from the events_by_id call.

        Args:
            data (dict): the part of the response from the ticketswitch
                `events_by_id` API call containing events and other data.

        Returns:
            :class:`Event <pyticketswitch.event.Event>`: a new
            :class:`Event <pyticketswitch.event.EVent>` object populated with
            the data from the API.

        """

        kwargs = cls.class_dict_from_api_data(data.get('event'))

        if data.get('add_ons'):
            addons=[
                Event.from_api_data(raw_addon)
                for raw_addon in data.get('add_ons')
            ]
            kwargs.update(addon_events=addons)

        if data.get('upsells'):
            upsells=[
                Event.from_api_data(raw_upsell)
                for raw_upsell in data.get('upsells')
            ]
            kwargs.update(upsell_events=upsells)

        if data.get('venue_is_enforced') is not None:
            kwargs.update(venue_is_enforced=data.get('venue_is_enforced'))

        return cls(**kwargs)

    def __repr__(self):
        return u'<Event {}:{}>'.format(
            self.id, self.description.encode('ascii', 'ignore'))


class EventMeta(PaginationMixin, CurrencyMeta):
    """Meta information about a set of events

    Attributes:
        currencies (dict): dictionary of
            :class:`Currency <pytickectswitch.currency.Currency>`) objects
            indexed on currency code.
        default_currency_code (str): unless other wise specified all prices in
            the related response will be in this currency.
        desired_currency_code (str):
            the currency that the user account is expecting. Useful for
            conversions.
        page_length (int): the number of items per page.
        page_number (int): the current page.
        pages_remaining (int): the number of pages remaining.
        results_remaining (int): the total number of remaining results after
            the current page.
        total_results (int): the total number of results.
    """
    pass



class PyticketswitchError(Exception):
    pass


class APIError(PyticketswitchError):
    def __init__(self, msg, code=None, response=None, *args, **kwargs):
        super(APIError, self).__init__(msg, *args, **kwargs)
        self.msg = msg
        self.code = code
        self.response = response


class AuthenticationError(APIError):
    pass


class InvalidParametersError(PyticketswitchError):
    pass


class InvalidResponseError(PyticketswitchError):
    pass


class InvalidGeoParameters(PyticketswitchError):
    pass


class IntegrityError(PyticketswitchError):

    def __init__(self, message, data, *args):
        super(IntegrityError, self).__init__(message, *args)
        self.data = data


class BackendError(PyticketswitchError):
    pass


class BackendBrokenError(BackendError):
    pass


class BackendDownError(BackendError):
    pass


class BackendThrottleError(BackendError):
    pass


class CallbackGoneError(APIError):
    pass


class OrderUnavailableError(PyticketswitchError):

    def __init__(self, msg, reservation=None, meta=None, *args, **kwargs):
        super(OrderUnavailableError, self).__init__(msg, *args, **kwargs)
        self.reservation = reservation
        self.meta = meta



from pyticketswitch.country import Country
from pyticketswitch.mixins import JSONMixin


class SendMethod(JSONMixin, object):
    """Describes a method of sending tickets to a customer.

    Attributes:
        code (str): identifier for the send method.
        cost (float): additional cost to the customer for this send method.
        description (str): human readable description of the send method.
        type (str): indicates the type of the send method.
        can_generate_self_print (bool): indicates if the send method can
            self print a voucher
        self_print_voucher_url (str): url for the client to download their
            voucher from
        final_comment (str): comment about dispatching tickets
    """

    def __init__(self, code, cost=None, description=None, typ=None,
                 permitted_countries=None, can_generate_self_print=False,
                 self_print_voucher_url=None, final_comment=None):
        self.code = code
        self.cost = cost
        self.description = description
        self.type = typ
        self.permitted_countries = permitted_countries
        self.can_generate_self_print = can_generate_self_print
        self.self_print_voucher_url = self_print_voucher_url
        self.final_comment = final_comment

    @classmethod
    def from_api_data(cls, data):
        """Creates a new SendMethod object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a send method.

        Returns:
            :class:`SendMethod <pyticketswitch.send_method.SendMethod>`: a new
            :class:`SendMethod <pyticketswitch.send_method.SendMethod>` object
            populated with the data from the api.

        """
        kwargs = {
            'code': data.get('send_code'),
            'description': data.get('send_desc'),
            'typ': data.get('send_type'),
            'can_generate_self_print': data.get('can_generate_self_print'),
            'self_print_voucher_url': data.get('self_print_voucher_url'),
            'final_comment': data.get('send_final_comment'),
        }

        cost = data.get('send_cost', 0.0)
        kwargs.update(cost=cost)

        permitted_countries_raw = data.get('permitted_countries', {})

        if permitted_countries_raw:
            permitted_countries_raw = permitted_countries_raw.get('country', [])

            permitted_countries = [
                Country.from_api_data(country)
                for country in permitted_countries_raw
            ]

            kwargs.update(permitted_countries=permitted_countries)

        return cls(**kwargs)

    def __repr__(self):
        return u'<SendMethod {}:{}>'.format(
            self.code, self.description.encode('ascii', 'ignore'))



from pyticketswitch import utils
from pyticketswitch.cost_range import CostRange
from pyticketswitch.availability import AvailabilityDetails
from pyticketswitch.mixins import JSONMixin, PaginationMixin
from pyticketswitch.currency import CurrencyMeta


class Performance(JSONMixin, object):
    """Describes and occurance of an :class:`Event <pyticketswitch.event.Event>`.

    The performance will have either a **date_time** or a **name**.

    .. note:: The **date_time** field is a timezone aware datetime and it is
              localised for the venue and not the user.

    Attributes:
        id (str): identifier for the performance.
        event_id (str): identifier for the event.
        date_time (:obj:`datetime.datetime`): the localised date and time for
            the performance.
        date_description (str): a human readable description of the date of the
            performance.
        time_description (str): a human readable description of the time of the
            performance.
        has_pool_seats (bool): the performance has pool seats available.
        is_limited (bool): the performance has limited availability.
        cached_max_seats (int): the maximum number of seats available to book
            in a single order. This value is cached and may not be accurate.
        cost_range (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            pricing summary, may also include offers.
        no_singles_cost_range (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            pricing summary when no leaving single seats, may also include
            offers.
        is_ghost (bool): the performance is a ghost performance and is nolonger
            available.
        name (str): the name of the performance.
        running_time (int): the number of minutes the performance is expected
            to run for.
        availability_details (:class:`AvailabilityDetails <pyticketswitch.availability.AvailabilityDetails>`):
            summerised availability data for the performance. This data is
            cached from previous availability calls and may not be accurate.

    """

    def __init__(self, id_, event_id, date_time=None,
                 date_description=None, time_description=None, has_pool_seats=False,
                 is_limited=False, cached_max_seats=None, cost_range=None,
                 no_singles_cost_range=None, is_ghost=False, name=None,
                 running_time=None, availability_details=None):

        self.id = id_
        self.event_id = event_id
        self.date_time = date_time
        self.date_description = date_description
        self.time_description = time_description
        self.has_pool_seats = has_pool_seats
        self.is_limited = is_limited
        self.cached_max_seats = cached_max_seats
        self.cost_range = cost_range
        self.no_singles_cost_range = no_singles_cost_range
        self.is_ghost = is_ghost
        self.name = name
        self.running_time = running_time
        self.availability_details = availability_details

    @classmethod
    def from_api_data(cls, data):
        """Creates a new **Performance** object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a performance.

        Returns:
            :class:`Performance <pyticketswitch.order.Performance>`: a new
            :class:`Performance <pyticketswitch.order.Performance>` object
            populated with the data from the api.

        """
        id_ = data.get('perf_id')
        event_id = data.get('event_id')

        date_time = data.get('iso8601_date_and_time')
        if date_time:
            date_time = utils.isostr_to_datetime(date_time)

        date_desc = data.get('date_desc')
        time_desc = data.get('time_desc')

        api_cost_range = data.get('cost_range', {})
        api_no_singles_cost_range = api_cost_range.get('no_singles_cost_range', {})
        cost_range = None
        no_singles_cost_range = None

        if api_cost_range:
            api_cost_range['singles'] = True
            cost_range = CostRange.from_api_data(api_cost_range)

        if api_no_singles_cost_range:
            api_no_singles_cost_range['singles'] = False
            no_singles_cost_range = CostRange.from_api_data(
                api_no_singles_cost_range)

        availability_details = AvailabilityDetails.from_api_data(
            data.get('avail_details', {}))

        kwargs = {
            'id_': id_,
            'event_id': event_id,
            'date_time': date_time,
            'date_description': date_desc,
            'time_description': time_desc,
            'running_time': data.get('running_time'),
            'name': data.get('perf_name'),
            'has_pool_seats': data.get('has_pool_seats', False),
            'is_limited': data.get('is_limited', False),
            'is_ghost': data.get('is_ghost', False),
            'cached_max_seats': data.get('cached_max_seats'),
            'cost_range': cost_range,
            'no_singles_cost_range': no_singles_cost_range,
            'availability_details': availability_details,
        }

        return cls(**kwargs)

    def __repr__(self):
        if self.date_time:
            return u'<Performance {}: {}>'.format(self.id, self.date_time.isoformat())
        return u'<Performance {}>'.format(self.id)


class PerformanceMeta(PaginationMixin, CurrencyMeta, object):
    """
    PerformanceMeta contains meta information about a list of
    :class:`Performances <pyticketswitch.performance.Performance>`.

    Attributes:
        auto_select (bool): indicates that the performance list will contain
            only one performance and this performance should be automatically
            selected for the customer. See :ref:`Events that don't need
            performance selection
            <events_that_dont_need_performance_selection>` for more
            information.
        has_names: (bool): indicates that the related performances have names
    """

    def __init__(self, auto_select=False, has_names=False, *args, **kwargs):
        self.auto_select = auto_select
        self.has_names = has_names
        super(PerformanceMeta, self).__init__(*args, **kwargs)

    @classmethod
    def from_api_data(cls, data):
        """Creates a new **PerformanceMeta** object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns some performance meta data.

        Returns:
            :class:`PerformanceMeta <pyticketswitch.order.PerformanceMeta>`: a
            new :class:`PerformanceMeta <pyticketswitch.order.PerformanceMeta>`
            object populated with the data from the api.

        """
        inst = super(PerformanceMeta, cls).from_api_data(data)
        inst.auto_select = data.get('autoselect_this_performance', False)
        inst.has_names = data.get('results', {}).get('has_perf_names', False)

        return inst



from pyticketswitch.mixins import JSONMixin


class Commission(JSONMixin, object):
    """Commission payable to the user on the sale of a specfic order.

    Attributes:
       including_vat (float): the amount payable including value added tax.
       excluding_vat (float): the amount payable excluding value added tax.
       currency_code (str): the currency code the commission is priced in.
    """

    def __init__(self, including_vat, excluding_vat, currency_code):
        self.including_vat = including_vat
        self.excluding_vat = excluding_vat
        self.currency_code = currency_code

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Commission object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a commission payable to the user.

        Returns:
            :class:`Commission <pyticketswitch.commission.Commission>`: a new
            :class:`Commission <pyticketswitch.commission.Commission>` object
            populated with the data from the api.

        """

        kwargs = {
            'including_vat': data.get('amount_including_vat'),
            'excluding_vat': data.get('amount_excluding_vat'),
            'currency_code': data.get('commission_currency_code'),
        }

        return cls(**kwargs)



from pyticketswitch.order import Order
from pyticketswitch.mixins import JSONMixin
from pyticketswitch.debitor import Debitor
from pyticketswitch.purchase_result import PurchaseResult


class Bundle(JSONMixin, object):
    """A collection of orders into the same backend system

    Attributes:
        source_code (str): the identifier of the backend system.
        orders (list): the orders.
        description (str): description of the backend system.
        total_seatprice (float): the total cost of the seats/tickets in the
            bundle.
        total_surcharge (float): the total additional charges for the bundle.
        total_send_cost (float): the total postage fee for the bundle.
        total (float): the total amount of money required to purchase this
            bundle.
        currency_code (str): the currency that the prices are in.
        debitor (:class:`Debitor <pyticketswitch.debitor.Debitor>`):
            information about an external debitor that will be used to take
            payment for this bundle. When you are selling on credit, or the
            source system is taking payment, this will be :obj:`None`.
        terms_and_conditions (str): supplier terms and conditions. Only
            availabile when requested with the optional ``source_info`` flag.
        purchase_result (:class:`PurchaseResult <pyticketswitch.purchase_result.
            PurchaseResult>`): the result of the purchase attempt for the bundle

    """

    def __init__(self, source_code, orders=None, description=None,
                 total_seatprice=None, total_surcharge=None,
                 total_send_cost=None, total=None, currency_code=None,
                 debitor=None, terms_and_conditions=None, purchase_result=None):
        self.source_code = source_code
        self.orders = orders
        self.description = description
        self.total_seatprice = total_seatprice
        self.total_surcharge = total_surcharge
        self.total_send_cost = total_send_cost
        self.total = total
        self.currency_code = currency_code
        self.debitor = debitor
        self.terms_and_conditions = terms_and_conditions
        self.purchase_result = purchase_result

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Bundle object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a bundle.

        Returns:
            :class:`Bundle <pyticketswitch.bundle.Bundle>`: a new
            :class:`Bundle <pyticketswitch.bundle.Bundle>` object
            populated with the data from the api.

        """

        kwargs = {
            'source_code': data.get('bundle_source_code'),
            'description': data.get('bundle_source_desc'),
            'currency_code': data.get('currency_code'),
            'terms_and_conditions': data.get('source_t_and_c'),
        }

        raw_orders = data.get('order')
        if raw_orders:
            orders = [Order.from_api_data(order) for order in raw_orders]
            kwargs.update(orders=orders)

        # Below we are explicital checking for not None because we want to
        # differentiate between situtations where a value is 0 and a value is
        # missing from the response.
        total_seatprice = data.get('bundle_total_seatprice')
        if total_seatprice is not None:
            kwargs.update(total_seatprice=total_seatprice)

        total_surcharge = data.get('bundle_total_surcharge')
        if total_surcharge is not None:
            kwargs.update(total_surcharge=total_surcharge)

        total_send_cost = data.get('bundle_total_send_cost')
        if total_send_cost is not None:
            kwargs.update(total_send_cost=total_send_cost)

        total = data.get('bundle_total_cost')
        if total is not None:
            kwargs.update(total=total)

        raw_debitor = data.get('debitor')
        if raw_debitor:
            debitor = Debitor.from_api_data(raw_debitor)
            kwargs.update(debitor=debitor)

        raw_purchase_result = data.get('purchase_result')
        if raw_purchase_result:
            purchase_result = PurchaseResult.from_api_data(raw_purchase_result)
            kwargs.update(purchase_result=purchase_result)

        return cls(**kwargs)

    def get_events(self):
        """Get the events in the bundle.

        Returns:
            list: list of :class:`Event <pyticketswitch.event.Event>` objects.
        """
        if not self.orders:
            return []
        return [
            order.event
            for order in self.orders
            if order.event and order.event.id
        ]

    def get_event_ids(self):
        """Get the event ids of the events in the bundle.

        Returns:
            set: set of events IDs (str).

        """
        return {event.id for event in self.get_events()}

    def is_purchased(self):
        """Check if this bundle was successfully purchased.

        Returns:
            bool: if the bundle is purchased or not.
        """
        if self.purchase_result:
            return self.purchase_result.success
        return False

    def __repr__(self):
        return u'<Bundle {}>'.format(self.source_code)



from collections import OrderedDict
from pyticketswitch.mixins import JSONMixin
from pyticketswitch.debitor import Debitor


class Integration(JSONMixin, object):
    """Information to use when integrating payment methods into clients.

    Attributes:
        type (str): the debitor type.
        amount (float): the amount to be debited.
        amount_base (int): the precise amount. see currency for places.
        currency (str): the currency code.
        data (dict): debitor specific information.

    """

    def __init__(self, typ, amount=None, base_amount=None, currency=None,
                 data=None):
        self.type = typ
        self.amount = amount
        self.base_amount = base_amount
        self.currency = currency
        self.data = data

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Integration object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns the integration.

        Returns:
            :class:`Integration <pyticketswitch.callout.Integration>`: a new
            :class:`Integration <pyticketswitch.callout.Integration>` object
            populated with the data from the api.

        """

        kwargs = {
            'typ': data.get('debitor_type'),
            'amount': data.get('debit_amount'),
            'base_amount': data.get('debit_base_amount'),
            'currency': data.get('debit_currency'),
            'data': data.get('debitor_specific_data'),
        }

        return cls(**kwargs)


class Callout(JSONMixin, object):
    """Information about a redirection required by a 3rd party the payment provider.

    Attributes:
        code (str): the identifier for the source system of the
            bundle this callout will be paying for.
        description: the human readable description for the source
            system.
        total (float): the total amount of money that will be payable.
        type (str): the HTTP method that should be used to send the customer
            to the destination.
        destination (str): the destination url to send your user to.
        parameters (:py:class:`collections.OrderedDict`): dictionary of key
            value pairs that the 3rd party is expecting as query string
            parameters or form parameters.
        html (str): a blob of HTML that can be rendered to produce either the
            required redirect or the form to continue with the transaction.
        integration_data (dict):
            data relevant to the debitor that the customer will be redirected
            to. This *may* provide a opertunity to integrate directly. Consult
            documentation for more details.
        debitor (:class:`Debitor <pyticketswitch.debitor.Debitor>`): the debitor
            information that can be used for a front end integration.
        return_token (str): return token specified in the last purchase or
            callback call.
    """

    def __init__(self, code=None, description=None, total=None, typ=None,
                 destination=None, parameters=None, integration_data=None,
                 debitor=None, currency_code=None, return_token=None):

        self.code = code
        self.description = description
        self.total = total
        self.type = typ
        self.destination = destination
        self.parameters = parameters
        self.integration_data = integration_data
        self.debitor = debitor
        self.currency_code = currency_code
        self.return_token = return_token

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Callout object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a callout.

        Returns:
            :class:`Callout <pyticketswitch.callout.Callout>`: a new
            :class:`Callout <pyticketswitch.callout.Callout>` object
            populated with the data from the api.

        """

        kwargs = {
            'code': data.get('bundle_source_code'),
            'description': data.get('bundle_source_desc'),
            'total': data.get('bundle_total_cost'),
            'destination': data.get('callout_destination_url'),
            'typ': data.get('callout_type'),
            'currency_code': data.get('currency_code'),
            'return_token': data.get('return_token'),
        }

        integration_data = data.get('callout_integration_data')

        if integration_data:
            kwargs.update(integration_data=integration_data)

        raw_debitor = data.get('debitor')
        if raw_debitor:
            debitor = Debitor.from_api_data(raw_debitor)
            kwargs.update(debitor=debitor)

        raw_parameters = data.get('callout_parameters')
        parameter_order = data.get('callout_parameters_order')

        if raw_parameters:
            if parameter_order:
                ordered_parameters = (
                    [key, raw_parameters[key]]
                    for key in parameter_order
                )
                parameters = OrderedDict(ordered_parameters)
            else:
                parameters = OrderedDict(raw_parameters.items())

            kwargs.update(parameters=parameters)

        return cls(**kwargs)

    def __repr__(self):
        return u'<Callout {}:{}>'.format(self.code, self.return_token)



from pyticketswitch.mixins import JSONMixin


class CardType(JSONMixin, object):
    """Type of card that will be accepted during the payment process

    Attributes:
        code (str): identifier for the card type.
        description (str): human readable description of the card type.

    """

    def __init__(self, code=None, description=None):
        self.code = code
        self.description = description

    def __repr__(self):
        return u'<CardType {}: {}>'.format(self.code, self.description)



import datetime
import decimal
import json


class JSONMixin(object):
    """Adds json encoding functionality to objects."""

    def __jsondict__(self, hide_none=True, hide_empty=True):

        def sanitise(obj):
            if isinstance(obj, datetime.datetime):
                return obj.isoformat()

            if isinstance(obj, datetime.date):
                return obj.isoformat()

            if isinstance(obj, decimal.Decimal):
                return float(obj)

            if hasattr(obj, '__jsondict__'):
                return obj.__jsondict__(hide_none=hide_none, hide_empty=hide_empty)

            if isinstance(obj, list):
                return [sanitise(item) for item in obj]

            if isinstance(obj, dict):
                return {key: sanitise(value) for key, value in obj.items()}

            return obj

        return {
            key: sanitise(obj)
            for key, obj in self.__dict__.items()

            # when hiding None's and the object is None, skip the object
            if not (hide_none and obj is None)

            # when hiding empty iterators and the object is an iterator and
            # it's empty, skip the object.
            if not (hide_empty and hasattr(obj, '__iter__') and not obj)
        }

    def as_dict_for_json(self, hide_none=True, hide_empty=True):
        """Generate a json serialisable dictionary from the object

        Dates are changed to strings in ISO 8601 format.

        Lists children are recursively serialised.

        Dictionary values are recursively serialised, keys are left as is.

        Objects that implement __jsondict__ return the result of calling
        __jsondict__ on the object.

        Args:
            hide_none (bool, optional): when :obj:`True` the returned
                dictionary will not include attributes who's value is
                :obj:`None`.
            hide_empty (bool, optional): when :obj:`True` the returned
                dictionary will not include attributes who's value is iterable
                and has a length of zero.

        Returns:
            dict: a dictionary representation of the object.

        """
        return self.__jsondict__(hide_none=hide_none, hide_empty=hide_empty)

    def as_json(self, hide_none=True, hide_empty=True, **kwargs):
        """Generate a json represetation of an object.

        Args:
            hide_none (bool): when :obj:`True` the returned
                dictionary will not include attributes who's value is
                :obj:`None`.
            hide_empty (bool): when :obj:`True` the returned
                dictionary will not include attributes who's value is iterable
                and has a length of zero.
            **kwargs: passed to :py:func:`json.dumps`.

        Returns:
            str: a json representation of an object

        """

        return json.dumps(
            self.as_dict_for_json(hide_none=hide_none, hide_empty=hide_empty),
            **kwargs
        )


class PaginationMixin(object):
    """Adds pagination information to a responses meta data.

    Attributes:
        page_length (int): the number of items per page.
        page_number (int): the current page.
        pages_remaining (int): the number of pages remaining.
        results_remaining (int): the total number of remaining results after
            the current page.
        total_results (int): the total number of results.

    """

    def __init__(self, page_length=None, page_number=None, pages_remaining=None,
                 total_results=None, results_remaining=None, *args, **kwargs):
        self.page_length = page_length
        self.page_number = page_number
        self.pages_remaining = pages_remaining
        self.results_remaining = results_remaining
        self.total_results = total_results
        super(PaginationMixin, self).__init__(*args, **kwargs)

    @classmethod
    def from_api_data(cls, data, result_key=None, *args, **kwargs):
        inst = super(PaginationMixin, cls).from_api_data(data)

        if not result_key:
            result_key = 'results'

        results = data.get(result_key, {})
        paging_data = results.get('paging_status', {})

        inst.page_length = paging_data.get('page_length')
        inst.page_number = paging_data.get('page_number')
        inst.pages_remaining = paging_data.get('pages_remaining')
        inst.results_remaining = paging_data.get('results_remaining')
        inst.total_results = paging_data.get('total_unpaged_results')

        return inst

    def is_paginated(self):
        """Indicates that the response is paginated

        Returns:
            bool: :obj:`True` when the response is omitting results due to
            pagination otherwise :obj:``False``

        """

        if not self.total_results or not self.page_length:
            return False

        if self.total_results < self.page_length:
            return False

        return True


class SeatPricingMixin(object):
    """Adds seat pricing to an object

    Attributes:
        seatprice (float): the price per seat/ticket.
        surcharge (float): additional charges per seat/ticket.
        non_offer_seatprice (float): the original price per seat/ticket when
            not on offer.
        non_offer_surcharge (float): the original additional charges per
            seat/ticket when not on offer.
    """

    def __init__(self, seatprice=None, surcharge=None, non_offer_seatprice=None,
                 non_offer_surcharge=None, *args, **kwargs):
        super(SeatPricingMixin, self).__init__(*args, **kwargs)
        self.seatprice = seatprice
        self.surcharge = surcharge
        self.non_offer_seatprice = non_offer_seatprice
        self.non_offer_surcharge = non_offer_surcharge

    @staticmethod
    def kwargs_from_api_data(data):
        kwargs = {
            'seatprice': data.get('sale_seatprice'),
            'surcharge': data.get('sale_surcharge'),
            'non_offer_seatprice': data.get('non_offer_sale_seatprice'),
            'non_offer_surcharge': data.get('non_offer_sale_surcharge'),
        }
        return kwargs

    def combined_price(self):
        """Returns the combined seatprice and surcharge.

        This method assumes that we have both a seatprice and surcharge.
        In the situation where are missing either a seatprice or a surcharge
        then we don't have all the information to be able provide this
        information.

        Returns:
            float: the combined seatprice and surcharge

        Raises:
            AssertionError: It might seem like the obvious thing to do would be
                to assume the missing data was in fact zero and simply allow the
                addition to continue. However that would be somewhat dangerous when
                we are talking about prices, and it's better to actually raise an
                exception to indicate that there was a problem with the objects
                data, than to inform a customer that the tickets are free or have
                no booking fees

        """
        assert self.seatprice is not None, 'seatprice data missing'
        assert self.surcharge is not None, 'surcharge data missing'
        return self.seatprice + self.surcharge

    def non_offer_combined_price(self):
        """Returns the combined non offer seatprice and surcharge.

        This method assumes that we have both a seatprice and surcharge.
        In the situation where are missing either a seatprice or a surcharge
        then we don't have all the information to be able provide this
        information.

        Returns:
            float: the combined seatprice and surcharge

        Raises:
            AssertionError: It might seem like the obvious thing to do would be
                to assume the missing data was in fact zero and simply allow the
                addition to continue. However that would be somewhat dangerous when
                we are talking about prices, and it's better to actually raise an
                exception to indicate that there was a problem with the objects
                data, than to inform a customer that the tickets are free or have
                no booking fees

        """
        assert self.non_offer_seatprice is not None, 'non_offer_seatprice data missing'
        assert self.non_offer_surcharge is not None, 'non_offer_surcharge data missing'
        return self.non_offer_seatprice + self.non_offer_surcharge



from pyticketswitch.offer import Offer
from pyticketswitch.mixins import JSONMixin
from pyticketswitch import utils


class CostRange(JSONMixin, object):
    """CostRange gives summarized pricing for events and performances.

    This information is returned from cached data collected when making actual
    calls to the backend system, and should not be considered accurate.

    Attributes:
        valid_quantities (list): list of valid quanities available for
            purchase.
        min_seatprice (float): the minimum cost per seat the customer might be
            expected to pay.
        max_seatprice (float): the maximum cost per seat the customer might be
            expected to pay.
        min_surcharge (float): the minimum surcharge per seat the customer
            might be expected to pay.
        max_surcharge (float): the maximum surcharge per seat the customer
            might be expected to pay.
        currency (str): currency the cost range and offer prices are in.
        best_value_offer (:class:`Offer <pyticketswitch.offer.Offer>`): offer
            with the highest percentage saving.
        max_saving_offer (:class:`Offer <pyticketswitch.offer.Offer>`): offer
            with the highest absolute saving.
        min_cost_offer (:class:`Offer <pyticketswitch.offer.Offer>`): offer
            with the lowest cost.
        top_price_offer (:class:`Offer <pyticketswitch.offer.Offer>`): offer
            with the top price.

    """

    def __init__(self, valid_quantities=None, max_surcharge=None, max_seatprice=None,
                 min_surcharge=None, min_seatprice=None, allows_singles=True,
                 currency=None, best_value_offer=None, max_saving_offer=None,
                 min_cost_offer=None, top_price_offer=None):

        self.valid_quantities = valid_quantities
        self.max_seatprice = max_seatprice
        self.max_surcharge = max_surcharge
        self.min_seatprice = min_seatprice
        self.min_surcharge = min_surcharge
        self.currency = currency
        self.best_value_offer = best_value_offer
        self.max_saving_offer = max_saving_offer
        self.min_cost_offer = min_cost_offer
        self.top_price_offer = top_price_offer

    @classmethod
    def from_api_data(cls, data):
        """Creates a new CostRange object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a cost range.

        Returns:
            :class:`CostRange <pyticketswitch.cost_range.CostRange>`: a new
            :class:`CostRange <pyticketswitch.cost_range.CostRange>` object
            populated with the data from the api.

        """

        min_seatprice = utils.get_price(data, 'min_seatprice')
        min_surcharge = utils.get_price(data, 'min_surcharge')
        max_seatprice = utils.get_price(data, 'max_seatprice')
        max_surcharge = utils.get_price(data, 'max_surcharge')

        kwargs = {
            'valid_quantities': data.get('valid_quantities'),
            'min_surcharge': min_surcharge,
            'min_seatprice': min_seatprice,
            'max_surcharge': max_surcharge,
            'max_seatprice': max_seatprice,
            'allows_singles': data.get('singles', True),
            'currency': data.get('range_currency_code'),
        }

        best_value_offer = data.get('best_value_offer')
        if best_value_offer:
            best_value_offer = Offer.from_api_data(best_value_offer)
            kwargs.update(best_value_offer=best_value_offer)

        max_saving_offer = data.get('max_saving_offer')
        if max_saving_offer:
            max_saving_offer = Offer.from_api_data(max_saving_offer)
            kwargs.update(max_saving_offer=max_saving_offer)

        min_cost_offer = data.get('min_cost_offer')
        if min_cost_offer:
            min_cost_offer = Offer.from_api_data(min_cost_offer)
            kwargs.update(min_cost_offer=min_cost_offer)

        top_price_offer = data.get('top_price_offer')
        if top_price_offer:
            top_price_offer = Offer.from_api_data(top_price_offer)
            kwargs.update(top_price_offer=top_price_offer)

        return cls(**kwargs)

    def has_offer(self):
        return any([
            self.best_value_offer,
            self.max_saving_offer,
            self.min_cost_offer,
            self.top_price_offer
        ])

    def get_min_combined_price(self):
        return self.min_surcharge + self.min_seatprice

    def get_max_combined_price(self):
        return self.max_surcharge + self.max_seatprice


class CostRangeDetails(JSONMixin, object):
    """Summarizes pricing by ticket types and price bands for an event/performance

    This information is returned from cached data collected when making actual
    calls to the backend system, and should not be considered accurate.

    Attributes:
        ticket_type_code (str): identifier for the ticket type.
        price_band_code (str): identifier for the price band.
        cost_range (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            summary of pricing distribution for this ticket type/price band.
        cost_range_no_singles (:class:`CostRange <pyticketswitch.cost_range.CostRange>`):
            summary of pricing distribution for this ticket type/price band
            when not leaving any singles.
        ticket_type_description (str): description of the ticket type.
        price_band_description (str): description of the price band.

    """

    def __init__(self, ticket_type_code, price_band_code, cost_range,
                 ticket_type_description=None, price_band_description=None,
                 cost_range_no_singles=None, **kwargs):

        self.ticket_type_code = ticket_type_code
        self.price_band_code = price_band_code
        self.cost_range = cost_range
        self.cost_range_no_singles = cost_range_no_singles
        self.ticket_type_description = ticket_type_description
        self.price_band_description = price_band_description

    @classmethod
    def from_api_data(cls, data):
        details = []
        for ticket_type in data.get('ticket_type', []):
            kwargs = {
                'ticket_type_code': ticket_type.get('ticket_type_code'),
                'ticket_type_description': ticket_type.get('ticket_type_desc'),
            }
            for price_band in ticket_type.get('price_band', []):
                kwargs.update(
                    price_band_code=price_band.get('price_band_code'),
                    price_band_description=price_band.get('price_band_desc'),
                )

                cost_range = price_band.get('cost_range')

                if not cost_range:
                    continue

                kwargs.update(
                    cost_range=CostRange.from_api_data(cost_range)
                )

                no_singles = cost_range.get('no_singles_cost_range')

                if no_singles:
                    kwargs.update(
                        cost_range_no_singles=CostRange.from_api_data(no_singles)
                    )

                details.append(cls(**kwargs))

        return details



from pyticketswitch.order import Order
from pyticketswitch.status import Status


class Reservation(Status):
    """Describes some tickets currently being held for purchase

    Attributes:
        status (str): the currency status of the transaction.
        reserved_at (datetime.datetime): the date and time when the transaction
            was reserved.
        purchased_at (datetime.datetime): the date and time when the transaction
            was purchased.
        trolley (:class:`Trolley <pyticketswitch.trolley.Trolley>`): the
            contents of the transactions trolley.
        external_sale_page (str): the page that was rendered to the customer
            after the transaction was completed. This is only available if
            it was passed into the API at purchase time.
        languages (list): list of IETF language tags relevant to the
            transaction.
        remote_site (str): the remote site the transaction was reserved and
            purchased under.
        reserve_user (:class:`User <pyticketswitch.user.User>`): the user that
            was used to reserve the transaction.
        unreserved_orders (list): list of :class:`Orders
            <pyticketswitch.order.Order`>` that failed to reserve.
        prefilled_address (:class:`Address <pyticketswitch.address.Address>`):
            some address information that should be used to prefill any
            customer address fields. This is primarily used on B2B accounts.
        needs_payment_card (bool): When :obj:`True` indicates that this
            reservation will require :class:`CardDetails
            <pyticketswitch.payment_methods.CardDetails>` as the payment method
            at purchase time.
        needs_email_address (bool): When :obj:`True` indicates that the
            customer needs to provide a valid email address at purchase time.
        needs_agent_reference (bool): When :obj:`True` indicates that an agent
            reference should be provided at purchase time.
        can_edit_address (bool): When :obj:`False` indicates that the prefilled
            customer address provided by **prefilled_address** should not be
            edited.
        allowed_countries (list): list of :class:`Countries
            <pyticketswitch.country.Country>` that are acceptable for the
            customers postal address to be from.
        minutes_left (float): the number of minutes left before a reservation
            expires.

    """

    def __init__(self, unreserved_orders=None,
                 input_contained_unavailable_order=False, *args, **kwargs):

        super(Reservation, self).__init__(*args, **kwargs)
        self.unreserved_orders = unreserved_orders
        self.input_contained_unavailable_order = input_contained_unavailable_order

    @classmethod
    def from_api_data(cls, data):
        """Creates a new **Reservation** object from ticketswitch API data.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a reservation.

        Returns:
            :class:`Reservation <pyticketswitch.order.Reservation>`: a new
            :class:`Reservation <pyticketswitch.order.Reservation>` object
            populated with the data from the api.

        """

        inst = super(Reservation, cls).from_api_data(data)

        unreserved_orders = []
        raw_unreserved_orders = data.get('unreserved_orders')
        if raw_unreserved_orders:
            unreserved_orders = [
                Order.from_api_data(order)
                for order in raw_unreserved_orders
            ]

        inst.unreserved_orders = unreserved_orders
        inst.input_contained_unavailable_order = data.get(
            'input_contained_unavailable_order', False)

        return inst

    def __repr__(self):
        if self.trolley and self.trolley.transaction_uuid:
            return u'<Reservation {}>'.format(self.trolley.transaction_uuid)
        return super(Reservation, self).__repr__()



from pyticketswitch.mixins import JSONMixin


class PurchaseResult(JSONMixin, object):
    """Describes the results of a purchase attempt

    Attributes:
        success (bool): indicates if the purchase was successful.
        failed_3d_secure (bool): indicates if the purchase failed 3d secure.
        failed_avs (bool): indicates if the purchase failed address checks.
        failed_cv_two (bool): indicates if the purchase credit card
            verification number.
        error (string): error code for backend purchase failure. These are
            non uniform as they can come from a varitiy of sources.
        failure_reason (string): description of the failure reason, if there
            is one. Not for display to customers.
        is_partial (bool): indicates if the success is only partial for multi-
            bundle transactions where one may succeed and the other fail.
        is_semi_credit (bool): Marks the purchase as provisionally completed but
            requiring invoicing before sale is finally confirmed.

    """

    def __init__(self, success=False, failed_3d_secure=False, failed_avs=False,
                 failed_cv_two=False, error=None, failure_reason=None,
                 is_partial=False, is_semi_credit=False):

        self.success = success
        self.failed_3d_secure = failed_3d_secure
        self.failed_avs = failed_avs
        self.failed_cv_two = failed_cv_two
        self.error = error
        self.failure_reason = failure_reason
        self.is_partial = is_partial
        self.is_semi_credit = is_semi_credit

    @classmethod
    def from_api_data(cls, data):
        """Creates a new PurchaseResults object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns the results of a purchase attempt.

        Returns:
            :class:`PurchaseResult <pyticketswitch.purchase_result.PurchaseResult>`:
            a new
            :class:`PurchaseResult <pyticketswitch.purchase_result.PurchaseResult>`
            object populated with the data from the api.

        """

        kwargs = {
            'success': data.get('success'),
            'failed_3d_secure': data.get('failed_3d_secure'),
            'failed_avs': data.get('failed_avs'),
            'failed_cv_two': data.get('failed_cv_two'),
            'error': data.get('purchase_error'),
            'failure_reason': data.get('failure_reason'),
            'is_partial': data.get('is_partial'),
            'is_semi_credit': data.get('is_semi_credit'),
        }

        return cls(**kwargs)



from pyticketswitch.mixins import JSONMixin


class Country(JSONMixin, object):
    """Represents a country

    Attributes:
        code (str):  ISO 3166-1 country code.
        description (str): a human readable name for the country.

    """

    def __init__(self, code, description=None):
        self.code = code
        self.description = description

    @classmethod
    def from_api_data(cls, data):
        """Creates a new Country object from API data from ticketswitch.

        Args:
            data (dict): the part of the response from a ticketswitch API call
                that concerns a country.

        Returns:
            :class:`Country <pyticketswitch.country.Country>`: a new
            :class:`Country <pyticketswitch.country.Country>` object
            populated with the data from the api.

        """

        kwargs = {
            'code': data.get('country_code'),
            'description': data.get('country_desc'),
        }

        return cls(**kwargs)

    def __repr__(self):
        return u'<Country {}:{}>'.format(
            self.code, self.description.encode('ascii', 'ignore'))



import functools
import warnings

from datetime import date, datetime
from dateutil import parser
from pyticketswitch.exceptions import InvalidParametersError


def date_range_str(start_date, end_date):
    """Convert a set of dates to string readable by the API

    Args:
        start_date (datetime.date): the start of the date range.
        end_date (datetime.date): the end of the date range.

    Returns:
        str: a date range in the format of "YYYYMMDD:YYYYMMDD".

        Missing either or both dates is acceptable and will return
        "YYYYMMDD:", ":YYYYMMDD", ":".

    Raises:
        InvalidParametersError: when a start_date or end_date is specified and
            it is not a datetime.date object.

    """
    if start_date and not isinstance(start_date, date):
        raise InvalidParametersError("start_date is not a datetime instance")
    if end_date and not isinstance(end_date, date):
        raise InvalidParametersError("end_date is not a datetime instance")

    if start_date:
        start_date = start_date.strftime('%Y%m%d')
    else:
        start_date = ''

    if end_date:
        end_date = end_date.strftime('%Y%m%d')
    else:
        end_date = ''

    if start_date or end_date:
        date_range = '{}:{}'.format(start_date, end_date)
    else:
        date_range = ''

    return date_range


def isostr_to_datetime(date_str):
    """Convert an iso datetime string to a :py:class:`datetime.datetime` object.

    Args:
        date_str (str): the string to convert.

    Returns:
        :py:class:`datetime.datetime`: the python representation of the date
            and time.

    Raises:
        ValueError: when the date_str is empty or None.

    """
    if not date_str:
        raise ValueError('{} is not a valid datetime string'.format(date_str))

    dt = parser.parse(date_str)
    return dt


def yyyymmdd_to_date(date_str):
    """Convert a YYYYMMDDD formated date to python :py:class:`datetime.date` object.

    Args:
        date_str (str): the string to convert.

    Returns:
        :py:class:`datetime.date`: the python representation of the date.

    Raises:
        ValueError: when the date_str is empty or None.
    """
    if not date_str:
        raise ValueError('{} is not a valid datetime string'.format(date_str))

    date = datetime.strptime(date_str, '%Y%m%d')
    if date:
        return date.date()


def specific_dates_from_api_data(dates):

    MONTHS = {
        'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4,
        'may': 5, 'jun': 6, 'jul': 7, 'aug': 8,
        'sep': 9, 'oct': 10, 'nov': 11, 'dec': 12,
    }

    return [
        date(int(year.split('_')[1]), MONTHS.get(month), int(day.split('_')[1]))
        for year, months in dates.items()
        if year.startswith('year_')
        for month, days in months.items()
        for day, valid in days.items()
        if valid is True
    ]


def bitmask_to_boolean_list(mask):
    """Convert a bitmask to boolean list

    Args:
        mask (int): the mask returned by the API

    Returns:
        list: list of booleans.

    """
    of_length = max(1, mask.bit_length())
    return [
        bool(mask >> i & 1)
        for i in range(of_length)
    ]


def bitmask_to_numbered_list(mask):
    """Convert a bitmask to a numbered list

    Args:
        mask (int): the mask returned by the API

    Returns:
        list: list of integers

    """
    if mask is None:
        return []

    return [
        i+1
        for i in range(mask.bit_length() + 1)
        if bool(mask >> i & 1)
    ]


def get_price(data, key):
    """Extracts a price as a float from some data

    Args:
        data (dict): the data containing the price
        key (str): the key of the target price.

    Returns:
        float: the price as a float.

        When the dictionary is missing the requested key, returns :obj:`None`
    """

    price = data.get(key)
    if price is not None:
        price = float(price)

    return price


def filter_none_from_parameters(params):
    """Removes parameters whos value is :obj:None

    Args:
        params (dict): dictionary of parameters to be passed to the API.

    Returns:
        dict: the original parameters with any parameters whos value was
        :obj:`None` removed.

    """
    return {
        key: value
        for key, value in params.items()
        if value is not None
    }


def deprecation_warning(message, stacklevel=2):
    warnings.simplefilter('always', DeprecationWarning)  # turn off filter
    warnings.warn(message,
                  category=DeprecationWarning,
                  stacklevel=stacklevel)
    warnings.simplefilter('default', DeprecationWarning)  # reset filter


def deprecated(func):
    """Mark a function as deprecated and raise a warning

    This decorator is used to mark functions as deprecated. It will result in
    a warning being emitted when the function is called.

    Args:
        func: the function to be wrapped

    Returns:
        the wrapped function that raises a warning
    """

    @functools.wraps(func)
    def wrapped_func(*args, **kwargs):
        deprecation_warning(
            "Call to deprecated function {}".format(func.__name__)
        )
        return func(*args, **kwargs)

    # Mark the function as deprecated
    wrapped_func.is_deprecated = True
    return wrapped_func
